#
# Copyright 2008-2021 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

"""
Recipe requirements
"""

import inspect
import numina.exceptions
import collections.abc
import contextlib
import warnings

from numina.exceptions import NoResultFound
import numina.types.datatype as dt
from .validator import as_list as deco_as_list
from .query import Ignore


class EntryHolder(object):
    def __init__(self, tipo, description, destination, optional,
                 default, choices=None, validation=True, alias=None):

        super(EntryHolder, self).__init__()

        if tipo is None:
            self.type = dt.NullType()
        elif tipo in [bool, str, int, float, complex, list]:
            self.type = dt.PlainPythonType(ref=tipo())
        elif inspect.isclass(tipo):
            self.type = tipo()
        else:
            self.type = tipo

        self.description = description
        self.optional = optional
        self.dest = destination
        self.default = default
        self.choices = choices
        self.validation = validation
        self.alias = alias

    def __get__(self, instance, owner):
        """Getter of the descriptor protocol."""
        if instance is None:
            return self
        else:
            if self.dest not in instance._numina_desc_val:
                instance._numina_desc_val[self.dest] = self.default_value()

            return instance._numina_desc_val[self.dest]

    def __set__(self, instance, value):
        """Setter of the descriptor protocol."""
        try:
            cval = self.convert(value)
            if self.choices and (cval not in self.choices):
                errmsg = f'{cval} not in {self.choices}'
                raise numina.exceptions.ValidationError(errmsg)
        except (ValueError, TypeError, numina.exceptions.ValidationError) as err:

            if len(err.args) == 0:
                errmsg = 'UNDEFINED ERROR'
                rem = ()
            else:
                errmsg = err.args[0]
                rem = err.args[1:]

            msg = f'"{self.dest}": {errmsg}'
            newargs = (msg, ) + rem
            err.args = newargs
            raise

        instance._numina_desc_val[self.dest] = cval

    def convert(self, val):
        return self.type.convert(val)

    def validate(self, val):
        if self.validation:
            return self.type.validate(val)
        return True

    def default_value(self):
        if self.default is not None:
            return self.convert(self.default)
        if self.type.internal_default is not None:
            return self.type.internal_default
        if self.optional:
            return None
        else:
            fmt = 'Required {0!r} of type {1!r} is not defined'
            msg = fmt.format(self.dest, self.type)
            raise ValueError(msg)


class Result(EntryHolder):
    """Result holder for RecipeResult."""

    def __init__(self, ptype, description="", validation=True,
                 destination=None, optional=False, default=None, choices=None):
        super(Result, self).__init__(
            ptype, description, destination=destination, optional=optional,
            default=default, choices=choices, validation=validation
        )

#        if not isinstance(self.type, DataProductType):
#            raise TypeError('type must be of class DataProduct')

    def __repr__(self):
        return f'Result(type={self.type!r}, dest={self.dest!r})'

    def convert(self, val):
        return self.type.convert_out(val)


class Product(Result):
    """Product holder for RecipeResult.

    .. deprecated:: 0.16
            `Product` is replaced by `Result`. It will
            be removed in 1.0
    """

    def __init__(self, ptype, description="", validation=True,
                 destination=None, optional=False, default=None, choices=None):
        super(Product, self).__init__(
            ptype,
            description=description,
            validation=validation,
            destination=destination,
            optional=optional,
            default=default,
            choices=choices
        )

        warnings.warn("The 'Product' class was renamed to 'Result'",
                      DeprecationWarning, stacklevel=2)

    def __repr__(self):
        return f'Product(type={self.type!r}, dest={self.dest!r})'


@contextlib.contextmanager
def tags_as_scalar(obsres):
    saved = obsres.tags
    if isinstance(obsres.tags, list):
        obsres.tags = obsres.tags[0]
    try:
        yield obsres
    finally:
        obsres.tags = saved


@contextlib.contextmanager
def tags_as_list(obsres):
    saved = obsres.tags
    if not isinstance(obsres.tags, list):
        obsres.tags = [obsres.tags]
    try:
        yield obsres
    finally:
        obsres.tags = saved


class Requirement(EntryHolder):
    """Requirement holder for RecipeInput.

    Parameters
    ----------
    rtype : :class:`~numina.types.datatype.DataType` or Type[DataType]
       Object or class repressenting the yype of the requirement,
       it must be a subclass of DataType
    description : str
       Description of the Requirement. The value is used
       by `numina show-recipes` to provide human-readable documentation.
    destination : str, optional
       Name of the field in the RecipeInput object. Overrides the value
       provided by the name of the Requirement variable
    optional : bool, optional
       If `False`, the builder of the RecipeInput must provide a value
       for this Parameter. If `True` (default), the builder can skip
       this Parameter and then the default in `value` is used.
    default : optional
        The value provided by the Requirement if the RecipeInput builder
        does not provide one.
    choices : list of values, optional
        The possible values of the inputs. Any other value will raise
        an exception
    alias : str, optional
        Alternative name of the field in the RecipeInput object.
    """

    def __init__(self, rtype, description, destination=None, optional=False,
                 default=None, choices=None, validation=True, query_opts=None,
                 alias=None):
        super(Requirement, self).__init__(
            rtype, description, destination=destination,
            optional=optional, default=default, choices=choices,
            validation=validation, alias=alias
        )

        self.query_opts = query_opts

        self.hidden = False

    def convert(self, val):
        return self.type.convert_in(val)

    def query_options(self, options=None):

        if options:
            effective_options = options
        else:
            effective_options = self.query_opts
        return effective_options

    def _check_dest_is_set(self):
        if self.dest is None:
            raise ValueError("destination value is not set, "
                             "use the constructor to set destination='value' "
                             "explicitly")

    def query(self, dal, obsres, options=None):
        from numina.core.query import ResultOf

        self._check_dest_is_set()

        q_options = self.query_options(options)

        if isinstance(q_options, Ignore):
            # we do not perform any query
            return self.default

        try:
            return self.query_on_ob(obsres)
        except NoResultFound:
            pass

        if isinstance(q_options, ResultOf):
            value = dal.search_result_relative(
                self.dest, self.type, obsres, result_desc=q_options
            )
            return value.content

        return self.query_on_dal(dal, obsres, options=q_options)

    def query_on_dal(self, dal, obsres, options=None):
        return self.query_on_dal_rec(self.type, dal, obsres, options=options)

    def query_on_dal_rec(self, this_type, dal, obsres, options=None):
        import numina.types.multitype as mt

        mtype = isinstance(this_type, mt.MultiType)
        next_type = this_type.node_type
        scalar = this_type.internal_scalar

        if mtype:
            failures = []
            for next_type in this_type.node_type:

                try:
                    result = self.query_on_dal_rec(
                        next_type, dal, obsres, options=options)
                    this_type._current = next_type
                    return result
                except NoResultFound as notfound:
                    failures.append((next_type, notfound))
            else:
                # Not found
                for subtype, notfound in failures:
                    pass  # subtype.on_query_not_found(notfound)
                raise NoResultFound

        if not scalar and this_type.multi_query:

            with tags_as_list(obsres) as obsres:
                query_tags = obsres.tags
                result = []
                for idx, tags in enumerate(query_tags):
                    obsres.tags = tags
                    res = self.query_on_dal_rec(
                        next_type, dal, obsres, options=options)
                    result.append(res)
                return result
        else:
            with tags_as_scalar(obsres) as obsres:
                result = self.query_on_dal_base(
                    this_type, dal, obsres, options=options)
                return result

    def query_on_dal_base(self, next_type, dal, obsres, options=None):
        if next_type.isproduct():
            value = dal.search_product(
                self.dest, next_type, obsres, options=options)
        else:
            value = dal.search_parameter(
                self.dest, next_type, obsres, options=options)
        return value.content

    def on_query_not_found(self, notfound):
        pass

    def query_on_ob(self, ob):
        self._check_dest_is_set()
        # First check if the requirement is embedded
        # in the observation result
        # It can in ob.requirements
        # or directly in the structure (as in GTC)

        # Check if key and alias are both defined and emit a warning
        dest_d = self.dest in ob.requirements
        alias_d = self.alias in ob.requirements

        key = self.dest
        if dest_d:
            if alias_d:
                msg = f"Both '{self.dest}' and alias '{self.alias}' are defined"
                warnings.warn(msg, RuntimeWarning)
        else:
            if alias_d:
                key = self.alias

        if alias_d or dest_d:
            content = ob.requirements[key]
            value = self.type._datatype_load(content)
            return value
        try:
            return getattr(ob, key)
        except AttributeError:
            raise NoResultFound("Requirement.query_on_ob")

    def __repr__(self):
        sclass = type(self).__name__
        fmt = ("%s(dest=%r, description='%s', "
               "default=%s, optional=%s, type=%s, choices=%r)"
               )
        return fmt % (sclass, self.dest, self.description, self.default,
                      self.optional, self.type, self.choices)

    def query_constraints(self):
        return self.type.query_constraints()

    def tag_names(self):
        return self.type.tag_names()


def _process_nelem(nlem):
    if nlem is None:
        return False, (None, None)
    if isinstance(nlem, int):
        return True, (nlem, nlem)
    if isinstance(nlem, str):
        if nlem == '*':
            return True, (0, None)
        if nlem == '+':
            return True, (1, None)

    raise ValueError(f'value {nlem} is invalid')


def _recursive_type(value, nmin=None, nmax=None, accept_scalar=True):
    if isinstance(value, (list, tuple)):
        # Continue with contents of list
        if len(value) == 0:
            next_ = dt.AnyType()
        else:
            next_ = value[0]
        final = _recursive_type(next_, accept_scalar=accept_scalar)
        return dt.ListOfType(final, nmin=nmin, nmax=nmax, accept_scalar=accept_scalar)
    elif isinstance(value, dict):
        return dt.PlainPythonType(value)
    elif isinstance(value, (bool, str, int, float, complex)):
        next_ = value
        return dt.PlainPythonType(next_)
    else:
        return value


class Parameter(Requirement):
    """The Recipe requires a plain Python type.

    Parameters
    ----------
    value : plain python type
       Default value of the parameter, the requested type is inferred from
       the type of value.
    description: str
       Description of the parameter. The value is used
       by `numina show-recipes` to provide human-readible documentation.
    destination: str, optional
       Name of the field in the RecipeInput object. Overrides the value
       provided by the name of the Parameter variable
    optional: bool, optional
       If `False`, the builder of the RecipeInput must provide a value
       for this Parameter. If `True` (default), the builder can skip
       this Parameter and then the default in `value` is used.
    choices: list of  plain python type, optional
        The possible values of the inputs. Any other value will raise
        an exception
    validator: callable, optional
        A custom validator for inputs
    accept_scalar: bool, optional
        If `True`, when `value` is a list, scalar value inputs are converted
        to list. If `False` (default), scalar values will raise an exception
        if `value` is a list
    as_list: bool, optional:
        If `True`, consider the internal type a list even if `value` is scalar
        Default is `False`
    nelem: str or int, optional:
        If nelem is '*', the list can contain any number of objects. If is '+',
        the list must contain at least 1 element. With a number, the list must
        contain that number of elements.
    alias : str, optional
        Alternative name of the field in the RecipeInput object.
    """

    def __init__(self, value, description, destination=None, optional=True,
                 choices=None, validation=True, validator=None,
                 accept_scalar=False,
                 as_list=False, nelem=None,
                 alias=None
                 ):

        if nelem is not None:
            decl_list, (nmin, nmax) = _process_nelem(nelem)
        elif as_list:
            decl_list = True
            nmin = nmax = None
        else:
            decl_list = False
            nmin = nmax = None

        is_scalar = not isinstance(value, collections.abc.Iterable)

        if is_scalar and decl_list:
            accept_scalar = True
            value = [value]

        if isinstance(value, (bool, str, int, float, complex, list)):
            default = value
        else:
            default = None

        if validator is None:
            self.custom_validator = None
        elif callable(validator):
            if as_list:
                self.custom_validator = deco_as_list(validator)
            else:
                self.custom_validator = validator
        else:
            raise TypeError('validator must be callable or None')

        mtype = _recursive_type(value,
                                nmin=nmin, nmax=nmax,
                                accept_scalar=accept_scalar
                                )

        super(Parameter, self).__init__(
            mtype, description, destination=destination,
            optional=optional, default=default,
            choices=choices, validation=validation,
            alias=alias
        )

    def convert(self, val):
        """Convert input values to type values."""
        pre = super(Parameter, self).convert(val)

        if self.custom_validator is not None:
            post = self.custom_validator(pre)
        else:
            post = pre
        return post

    def validate(self, val):
        """Validate values according to the requirement"""
        if self.validation:
            self.type.validate(val)

            if self.custom_validator is not None:
                self.custom_validator(val)

        return True



# Copyright 2008-2023 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#


import warnings

from numina.util.convert import convert_qc

from .base import DataTypeBase
from .datatype import DataType
from .dataframe import DataFrame
from .qc import QC


class DataProductMixin(DataTypeBase):
    """A type that is a data product."""

    def __init__(self, *args, **kwds):
        super(DataProductMixin, self).__init__(*args, **kwds)
        self.quality_control = QC.UNKNOWN

    def generators(self):
        return []

    @classmethod
    def isproduct(cls):
        return True

    def name(self):
        """Unique name of the datatype"""
        sclass = type(self).__name__
        return f"{sclass}"

    def query_constraints(self):
        import numina.core.query
        return numina.core.query.Constraint()

    def extract_db_info(self, obj, keys):
        """Extract metadata from serialized file"""
        import astropy.io.fits as fits
        result = {}
        if isinstance(obj, dict):
            try:
                qc = obj['quality_control']
            except KeyError:
                qc = QC.UNKNOWN
        elif isinstance(obj, DataFrame):
            with obj.open() as hdulist:
                qc = self.datamodel.get_quality_control(hdulist)
        elif isinstance(obj, fits.HDUList):
            qc = self.datamodel.get_quality_control(obj)
        else:
            qc = QC.UNKNOWN

        result['quality_control'] = qc

        other = super(DataProductMixin, self).extract_db_info(obj, keys)
        result.update(other)
        return result

    def update_meta_info(self):
        result = super(DataProductMixin, self).update_meta_info()
        result['quality_control'] = self.quality_control.name
        return result

    def __getstate__(self):
        st = {}
        st['quality_control'] = self.quality_control

        other = super(DataProductMixin, self).__getstate__()
        st.update(other)
        return st

    def __setstate__(self, state):
        qcval = state.get('quality_control', 'UNKNOWN')
        self.quality_control = convert_qc(qcval)

        super(DataProductMixin, self).__setstate__(state)


class DataProductTag(DataProductMixin):
    """A type that is a data product.

    .. deprecated:: 0.16
            `DataProductTag` is replaced by `DataProductMixin`. It will
            be removed in 1.0

    """

    def __init__(self, *args, **kwargs):
        warnings.warn("The 'DataProductTag' class was renamed to 'DataProductMixin'",
                      DeprecationWarning)
        super(DataProductTag, self).__init__(*args, **kwargs)


class DataProductType(DataProductMixin, DataType):
    def __init__(self, ptype, default=None):
        super(DataProductType, self).__init__(ptype, default=default)


class ConfigurationTag(object):
    """A type that is part of the instrument configuration."""

    @classmethod
    def isconfiguration(cls):
        return True


if __name__ == '__main__':

    import numina.util.namespace as nm
    import numina.core.tagexpr as tagexpr

    class TagsNamespace(nm.Namespace):

        @staticmethod
        def p_(name):
            return tagexpr.Placeholder(name)

    class Tagged(object):

        def __init__(self, tags_ids):
            self.tag_ids = tags_ids
            tg = {key: tagexpr.TagRepr(key) for key in self.tag_ids}
            self.tags = TagsNamespace(**tg)
            self.tags_dict = tg



#
# Copyright 2024-2025 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

"""Define a 3D WCS"""

from astropy.coordinates import SkyCoord
from astropy.io import fits
import astropy.units as u
from astropy.wcs import WCS
import numpy as np


def define_3d_wcs(naxis1_ifu, naxis2_ifu, skycoord_center, spatial_scale, wv_lincal, instrument_pa, verbose):
    """Define a 3D WCS.

    Parameters
    ----------
    naxis1_ifu : `~astropy.units.Quantity`
        NAXIS1 value of the WCS object (along slice).
    naxis2_ifu : `~astropy.units.Quantity`
        NAXIS2 value of the WCS object (perpendicular to the slice).
    skycoord_center : `~astropy.coordinates.sky_coordinate.SkyCoord`
        Coordinates at the center of the detector.
    spatial_scale : `~astropy.units.Quantity`
        Spatial scale per pixel.
    wv_lincal : `~.linear_wavelength_calibration.LinearWaveCal`
        Linear wavelength calibration object.
    instrument_pa : `~astropy.units.Quantity`
        Instrument Position Angle.
    verbose : bool
        If True, display additional information.

    Returns
    -------
    wcs3d : `~astropy.wcs.wcs.WCS`
        WCS of the data cube.

    """

    # initial checks
    if not naxis1_ifu.unit.is_equivalent(u.pix):
        raise ValueError(f'Unexpected naxis1 unit: {naxis1_ifu.unit}')
    if not naxis2_ifu.unit.is_equivalent(u.pix):
        raise ValueError(f'Unexpected naxis2 unit: {naxis2_ifu.unit}')
    if not isinstance(skycoord_center, SkyCoord):
        raise ValueError(f'Expected SkyCoord instance not found: {skycoord_center} of type {type(skycoord_center)}')
    if not spatial_scale.unit.is_equivalent(u.deg / u.pix):
        raise ValueError(f'Unexpected spatial_scale unit: {spatial_scale.unit}')

    default_wv_unit = wv_lincal.default_wavelength_unit

    # Define FITS header
    # Note: when using CDi_j keywords, there is no need to define CDELTi
    #       keywords, since they are ignored by astropy.wcs.WCS.
    #
    #       The FITS standard defined 
    #       in https://fits.gsfc.nasa.gov/standard40/fits_standard40aa-le.pdf 
    #       (page 30) states that there are three possible conventions:
    #       1. The CDi_j keywords: represent a full linear transformation matrix, 
    #          including rotation, scaling, and possible distortion. Used when you 
    #          want to describe the complete transformation in a single matrix.
    #       2. The PCi_j + CDELTi keywords: PCi_j should represent a rotation 
    #          and distortion matrix without units, meaning it should be normalized.
    #          The CDELTi keywords should represent the pixel scale in the
    #          corresponding axis, with units.
    #       3. CDELTi + ROTA2 keywords (not recommended anymore).
    #       A potential problem with the returned wcs3d object is that
    #       when using wcs3d.to_header() the CDi_j keywords are replaced by
    #       PCi_j and CDELTi keywords. This is actually a problem because
    #       the PCi_j keywords are not defined in the FITS standard (the
    #       resulting matrix should be normalized, which is not; see
    #       https://github.com/astropy/astropy/issues/1084 for more details).
    #       A possible solution is to define a custom wcs_to_header_using_cd_keywords()
    #       function that returns the header with CDi_j keywords (see below).
    #       This function should be used instead of wcs3d.to_header().
    header = fits.Header()
    header['NAXIS'] = 3
    header['NAXIS1'] = int(naxis1_ifu.value)
    header['NAXIS2'] = int(naxis2_ifu.value)
    header['NAXIS3'] = int(wv_lincal.naxis1_wavecal.value)
    header['CTYPE1'] = 'RA---TAN'
    header['CTYPE2'] = 'DEC--TAN'
    header['CTYPE3'] = 'WAVE'
    header['CRVAL1'] = skycoord_center.ra.deg
    header['CRVAL2'] = skycoord_center.dec.deg
    header['CRVAL3'] = wv_lincal.crval1_wavecal.to(default_wv_unit).value
    header['CRPIX1'] = (naxis1_ifu.value + 1) / 2
    header['CRPIX2'] = (naxis2_ifu.value + 1) / 2
    header['CRPIX3'] = wv_lincal.crpix1_wavecal.value
    spatial_scale_deg_pix = spatial_scale.to(u.deg / u.pix).value
    header['CD1_1'] = (-spatial_scale_deg_pix) * np.cos(instrument_pa).value
    header['CD1_2'] = spatial_scale_deg_pix * np.sin(instrument_pa).value
    header['CD2_1'] = (-spatial_scale_deg_pix) * (-np.sin(instrument_pa).value)
    header['CD2_2'] = spatial_scale_deg_pix * np.cos(instrument_pa).value
    header['CD3_3'] = wv_lincal.cdelt1_wavecal.to(default_wv_unit / u.pix).value
    header['CUNIT1'] = 'deg'
    header['CUNIT2'] = 'deg'
    header['CUNIT3'] = default_wv_unit.name

    # define wcs object
    wcs3d = WCS(header)
    if verbose:
        print(f'\n{wcs3d}')

    return wcs3d


def wcs_to_header_using_cd_keywords(wcs):
    """Return WCS header using CDi_j keywords.

    This function is a workaround to avoid the problem with the
    astropy.wcs.WCS.to_header() method that replaces CDi_j keywords
    with PCi_j and CDELTi keywords.

    Parameters
    ----------
    wcs : `~astropy.wcs.wcs.WCS`
        WCS object.

    Returns
    -------
    header : `~astropy.io.fits.header.Header`
        FITS header with CDi_j keywords.
    """
    header = wcs.to_header()

    # Replace PCi_j with CDi_j keywords and remove CDELTi keywords.
    list_of_pc_keys = [key for key in header.keys() if key.startswith('PC')]
    list_of_pc_keys.sort()  # Sort list of keys to ensure consistent order
    if len(list_of_pc_keys) > 0:
        for key in list_of_pc_keys:
            header.rename_keyword(key, f'CD{key[2]}_{key[4]}')
        list_of_cdelt_keys = [key for key in header.keys() if key.startswith('CDELT')]
        if len(list_of_cdelt_keys) > 0:
            for key in list_of_cdelt_keys:
                del header[key]
    else:
        # If no PCi_j keywords are present, rename CDELTi as CDi_i keywords
        list_of_cdelt_keys = [key for key in header.keys() if key.startswith('CDELT')]
        if len(list_of_cdelt_keys) > 0:
            for key in list_of_cdelt_keys:
                header.rename_keyword(key, f'CD{key[5]}_{key[5]}')
                header.comments[f'CD{key[5]}_{key[5]}'] = 'Coordinate transformation matrix element'
    return header


def get_wvparam_from_wcs3d(wcs3d):
    """Return CUNIT1, CRPIX1, CRVAL1, CDELT1 from 3D WCS.

    It is assumed that the wavelength axis corresponds to NAXIS3.

    Parameters
    ----------
    wcs3d : `~astropy.wcs.wcs.WCS`
        WCS of the data cube.

    Returns
    -------
    wv_cunit1 : `~astropy.units.core.Unit`
        Default wavelength unit to be employed in CRVAL1 and CDELT1.
    wv_crpix1 :`~astropy.units.Quantity`
        CRPIX1 value.
    wv_crval1 :`~astropy.units.Quantity`
        CRVAL1 value.
    wv_cdelt1 :`~astropy.units.Quantity`
        CDELT1 value.
    """

    wv_cunit1 = wcs3d.wcs.cunit[2]
    wv_crpix1 = wcs3d.wcs.crpix[2] * u.pix
    wv_crval1 = wcs3d.wcs.crval[2] * wv_cunit1

    # Note that the use of wcs3d.wcs.cdelt1[2] raises a RuntimeWarning:
    # "cdelt will be ignored since cd is present"
    # For that reason we read cdelt1 from the cd matrix
    wv_cdelt1 = wcs3d.wcs.cd[2, 2] * wv_cunit1 / u.pix

    return wv_cunit1, wv_crpix1, wv_crval1, wv_cdelt1


def header3d_after_merging_wcs2d_celestial_and_wcs1d_spectral(wcs2d_celestial, wcs1d_spectral):
    """Merge 2D celestial WCS and 1D spectral WCS into a single header.

    This function merges a 2D celestial WCS and a 1D spectral WCS into
    a single 3D WCS header. The celestial WCS is assumed to be in the 
    first two axes (NAXIS1 and NAXIS2), and the spectral WCS is assumed 
    to be in the third axis (NAXIS3). The resulting header contains the 
    necessary keywords to describe the 3D WCS, including the celestial 
    coordinates and the spectral wavelength information.

    The returned value is the FITS header and not a WCS object, 
    so to avoid astropy.wcs.WCS replacing the CDi_j keywords 
    with PCi_j and CDELTi keywords.

    Parameters
    ----------
    wcs2d_celestial : `~astropy.wcs.wcs.WCS`
        2D celestial WCS.
    wcs1d_spectral : `~astropy.wcs.wcs.WCS`
        1D spectral WCS.

    Returns
    -------
    header3d : `~astropy.io.fits.header.Header`
        FITS header with the merged WCS information.
    """

    header3d = wcs_to_header_using_cd_keywords(wcs2d_celestial)
    header3d['NAXIS'] = 3
    header_spectral = wcs_to_header_using_cd_keywords(wcs1d_spectral)
    header3d['WCSAXES'] = 3
    for item in ['CRPIX', 'CD', 'CUNIT', 'CTYPE', 'CRVAL']:
        # insert {item}3 after {item}2 to preserve the order in the header
        if item == 'CD':
            keybefore = f'{item}2_2'
            keyafter = f'{item}3_3'
            keyvalue = header_spectral[f'{item}1_1']
            keycomment = header_spectral.comments[f'{item}1_1']
        else:
            keybefore = f'{item}2'
            keyafter = f'{item}3'
            keyvalue = header_spectral[f'{item}1']
            keycomment = header_spectral.comments[f'{item}1']
        header3d.insert(
            keybefore,
            (keyafter, keyvalue, keycomment),
            after=True
        )

    return header3d



#
# Copyright 2008-2023 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#


"""Exceptions for the numina package."""


import numina.ext.gtc


class Error(Exception):
    """Base class for exceptions in the numina package."""
    pass


class RecipeError(Error):
    """A non recoverable problem during recipe execution."""
    pass


class DetectorElapseError(Error):
    """Error in the clocking of a Detector."""
    pass


class DetectorReadoutError(Error):
    """Error in the readout of a Detector."""
    pass


class ValidationError(Exception):
    """Error during validation of Recipe inputs and outputs."""
    pass


class NoResultFound(Exception):
    """No result found in a DAL query."""
    pass


NoResultFoundOrig = NoResultFound


# If we are in the GCS environment, use its exceptions
# where applies
if numina.ext.gtc.check_gtc():
    import gtc.SSL.GCSTypes

    # Save original exception
    NoResultFoundOrig = NoResultFound
    NoResultFound = gtc.SSL.GCSTypes.NotFound




import datetime
import pytest

from ..convert import convert_date


def test_convert_date1():
    result = datetime.datetime(2018, 2, 12, 22, 4, 30, 100000)
    value = convert_date("2018-02-12T22:04:30.1")
    assert value == result


def test_convert_date2():
    result = datetime.datetime(2018, 2, 12, 22, 4, 30)
    value = convert_date("2018-02-12T22:04:30")
    assert value == result


def test_convert_date3():
    result = datetime.datetime(2017, 6, 27, 0, 0)
    value = convert_date("2017-06-27")
    assert value == result


def test_convert_date4():
    with pytest.raises(ValueError):
        convert_date("something")



#
# Copyright 2008-2018 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

import unittest

import numpy

from numina.array.nirproc import fowler_array

MASK_GOOD = 0
MASK_SATURATION = 3


class FowlerTestCase(unittest.TestCase):
    def setUp(self):
        self.fdata = numpy.empty((10, 1, 1))
        rows = 3
        columns = 4
        self.emptybp = numpy.zeros((rows, columns), dtype='uint8')
        self.data = numpy.arange(10, dtype='int32')
        self.data = numpy.tile(self.data, (columns, rows, 1)).T
        self.blank = 1
        self.saturation = 65536
        self.sgain = 1.0
        self.sron = 1.0

    def test_exception(self):
        """Test we raise exceptions for invalid inputs in Fowler mode."""

        # Dimension must be 3
        self.assertRaises(ValueError, fowler_array,
                          numpy.empty((2,)))
        self.assertRaises(ValueError, fowler_array,
                          numpy.empty((2, 2)))
        self.assertRaises(ValueError, fowler_array,
                          numpy.empty((2, 3, 4, 5)))
        # saturation in good shape
        self.assertRaises(ValueError, fowler_array,
                          self.fdata, saturation=-100)
        self.assertRaises(ValueError, fowler_array,
                          self.fdata, saturation=0)
        # 0-axis must be even
        self.assertRaises(ValueError, fowler_array,
                          numpy.empty((5, 2, 0)))
        # gain must be positive
        self.assertRaises(ValueError, fowler_array,
                          self.fdata, gain=-1.0)
        self.assertRaises(ValueError, fowler_array,
                          self.fdata, gain=0)
        # RON must be positive
        self.assertRaises(ValueError, fowler_array,
                          self.fdata, ron=-1.0)
        self.assertRaises(ValueError, fowler_array,
                          self.fdata, ron=0)

    def test_saturation0(self):
        """Test we count correctly saturated pixels in Fowler mode."""

        # No points
        self.data[:] = 50000  # - 32768
        saturation = 40000  # - 32767

        res = fowler_array(self.data, saturation=saturation, blank=self.blank)

        # Number of points
        for nn in res[2].flat:
            self.assertEqual(nn, 0)

        # Mask value
        for n in res[3].flat:
            self.assertEqual(n, MASK_SATURATION)

        # Variance
        for v in res[1].flat:
            self.assertEqual(v, self.blank)

        # Values
        for v in res[0].flat:
            self.assertEqual(v, self.blank)

    def test_saturation1(self):
        """Test we count correctly saturated pixels in Fowler mode."""

        saturation = 50000
        self.data[7:, ...] = saturation

        res = fowler_array(self.data,
                           saturation=saturation,
                           blank=self.blank)

        for nn in res[2].flat:
            self.assertEqual(nn, 2)

        for n in res[3].flat:
            self.assertEqual(n, MASK_GOOD)

        for v in res[1].flat:
            self.assertAlmostEqual(v, 2.0/2)

        for v in res[0].flat:
            self.assertAlmostEqual(v, 5)

    def test_dtypes0(self):
        """Test output is float64 by default"""
        inttypes = ['int8', 'int16', 'int32', 'uint8', 'uint16', 'uint32']
        floattypes = ['float32', 'float64', ]
        if hasattr(numpy, 'float128'):
            floattypes.append('float128')
        mdtype = numpy.dtype('uint8')
        ddtype = numpy.dtype('float64')
        rows = 3
        columns = 4
        for dtype in inttypes:
            data = numpy.zeros(10, dtype=dtype)
            data[5:] = 10
            data = numpy.tile(self.data, (rows, columns, 1))
            res = fowler_array(data)
            self.assertIs(res[0].dtype, ddtype)
            self.assertIs(res[1].dtype, ddtype)
            self.assertIs(res[2].dtype, mdtype)
            self.assertIs(res[3].dtype, mdtype)

        for dtype in floattypes:
            data = numpy.zeros(10, dtype=dtype)
            data[5:] = 10
            data = numpy.tile(self.data, (rows, columns, 1))
            res = fowler_array(data)
            self.assertIs(res[0].dtype, ddtype)
            self.assertIs(res[1].dtype, ddtype)
            self.assertIs(res[2].dtype, mdtype)
            self.assertIs(res[3].dtype, mdtype)

    def test_badpixel0(self):
        """Test we ignore badpixels in Fowler mode."""
        mask_val = 2
        self.emptybp[...] = mask_val

        res = fowler_array(self.data,
                           saturation=self.saturation,
                           badpixels=self.emptybp,
                           blank=self.blank)

        for nn in res[2].flat:
            self.assertEqual(nn, 0)

        for n in res[3].flat:
            self.assertEqual(n, mask_val)

        for v in res[1].flat:
            self.assertEqual(v, self.blank)

        for v in res[0].flat:
            self.assertAlmostEqual(v, self.blank)

    def test_badpixel1(self):
        """Test we handle correctly None badpixel mask."""
        self.emptybp[...] = 0
        values = [2343, 2454, 2578, 2661, 2709, 24311, 24445,
                  24405, 24612, 24707]
        self.data = numpy.empty((10, 3, 4), dtype='int32')
        for i in range(10):
            self.data[i, ...] = values[i]
        arr = self.data[5:, 0, 0] - self.data[:5, 0, 0]
        mean = arr.mean()

        res = fowler_array(self.data,
                           saturation=self.saturation,
                           badpixels=self.emptybp,
                           blank=self.blank)

        for nn in res[2].flat:
            self.assertEqual(nn, 5)

        for n in res[3].flat:
            self.assertEqual(n, 0)

        # We have ti = texp = ts =0
        # So the noise is 2*sigma / N_p
        for v in res[1].flat:
            self.assertAlmostEqual(v, 2.0 / 5)

        for v in res[0].flat:
            self.assertAlmostEqual(v, mean)

        self.emptybp = None
        res = fowler_array(self.data,
                           saturation=self.saturation,
                           badpixels=self.emptybp,
                           blank=self.blank)

        for nn in res[2].flat:
            self.assertEqual(nn, 5)

        for n in res[3].flat:
            self.assertEqual(n, 0)

        for v in res[1].flat:
            self.assertAlmostEqual(v, 2.0/5)

        for v in res[0].flat:
            self.assertAlmostEqual(v, mean)

        res = fowler_array(self.data,
                           saturation=self.saturation,
                           blank=self.blank)

        for nn in res[2].flat:
            self.assertEqual(nn, 5)

        for n in res[3].flat:
            self.assertEqual(n, 0)

        for v in res[1].flat:
            self.assertAlmostEqual(v, 2.0/5)

        for v in res[0].flat:
            self.assertAlmostEqual(v, mean)

    def test_badpixel2(self):
        """Test we don't accept badpixel mask with incompatible shape."""
        self.assertRaises(ValueError, fowler_array, self.fdata,
                          badpixels=numpy.empty((10, 10)))
        self.assertRaises(ValueError, fowler_array, self.fdata,
                          badpixels=numpy.empty((1, 1, 1)))

    def test_badpixel3(self):
        """Test we don't accept badpixel mask with incompatible dtype."""
        self.assertRaises(ValueError, fowler_array, self.fdata,
                          badpixels=numpy.empty((1, 1), dtype='int'))

    def test_results1(self):
        """Test we obtain correct values in Fowler mode"""

        data = numpy.zeros((10, 4, 5), dtype='int32')
        vals = numpy.array([10, 13, 15, 17, 20, 411, 412, 414, 417, 422])
        ovals = vals[5:] - vals[:5]
        mean = ovals.mean()
        # var = ovals.var()

        for i in range(10):
            data[i, ...] = vals[i]

        res = fowler_array(data,
                           saturation=self.saturation,
                           blank=self.blank)

        for nn in res[2].flat:
            self.assertEqual(nn, 5)

        for n in res[3].flat:
            self.assertEqual(n, 0)

        for v in res[1].flat:
            self.assertAlmostEqual(v, 2.0 / 5)

        for v in res[0].flat:
            self.assertAlmostEqual(v, mean)



#
# Copyright 2008-2015 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

"""
Base metaclasses
"""

from .dataholders import Result
from .requirements import Requirement
import weakref


class StoreType(type):
    """Metaclass for storing members."""
    def __new__(cls, classname, parents, attributes):

        n_stored = weakref.WeakValueDictionary()
        for p in parents:
            stored = getattr(p, '__numina_stored__', None)
            if stored:
                n_stored.update(stored)

        new_attributes = {}
        for name, val in attributes.items():
            if cls.exclude(name, val):
                new_name, new_val = cls.transform(name, val)
                n_stored[new_name] = new_val
                new_attributes[new_name] = new_val
            else:
                new_attributes[name] = val

        new_attributes['__numina_stored__'] = n_stored

        return super(StoreType, cls).__new__(cls, classname, parents, new_attributes)

    def __setattr__(self, key, value):
        """Define __setattr__ in 'classes' created with this metaclass."""
        self._add_attr(key, value)

    def _add_attr(self, key, value):
        if self.exclude(key, value):
            nkey, nvalue = self.transform(key, value)
            self.__numina_stored__[nkey] = nvalue

        super(StoreType, self).__setattr__(key, value)

    @classmethod
    def exclude(cls, name, value):
        return False

    @classmethod
    def transform(cls, name, value):
        if value.dest is None:
            value.dest = name
        nname = value.dest
        return nname, value


class RecipeInputType(StoreType):
    """Metaclass for RecipeInput."""
    @classmethod
    def exclude(cls, name, value):
        return isinstance(value, Requirement)


class RecipeResultType(StoreType):
    """Metaclass for RecipeResult."""
    @classmethod
    def exclude(cls, name, value):
        return isinstance(value, Result)



from ..configorigin import UndefinedOrigin, ElementOrigin

import pytest


def test_undefined_origin():
    origin = UndefinedOrigin()
    assert origin.is_valid_date(None)


def _origin1():
    date_start = "2001-10-10T12:00:00"
    date_end = None
    name = "origin"
    cuuid = "c27287cd-90d1-417c-8238-f5ff220f818d"
    origin = ElementOrigin(name, cuuid, date_start=date_start, date_end=date_end)
    return origin


def _origin2():
    date_start = "2001-10-10T12:00:00"
    date_end = "2002-10-10T12:00:00"
    name = "origin"
    cuuid = "c27287cd-90d1-417c-8238-f5ff220f818d"
    origin = ElementOrigin(name, cuuid, date_start=date_start, date_end=date_end)
    return origin


_test_vals1 = [
    (None, True),
    ("2001-10-09T12:00:00", False),
    ("2001-10-11T12:00:00", True),
    ("2002-10-11T12:00:00", True),
    ("2002-10-10T12:00:00", True),
    ("2005-10-10T12:00:00", True),
]

_test_vals2 = [
    (None, True),
    ("2001-10-09T12:00:00", False),
    ("2001-10-10T12:00:00", True),
    ("2002-02-11T12:00:00", True),
    ("2002-10-10T12:00:00", False),
    ("2005-10-10T12:00:00", False),
]


@pytest.mark.parametrize("origin", [_origin1()])
@pytest.mark.parametrize("date, result", _test_vals1)
def test_origin1(origin, date, result):
    calc = origin.is_valid_date(date)
    assert result == calc


@pytest.mark.parametrize("origin", [_origin2()])
@pytest.mark.parametrize("date, result", _test_vals2)
def test_origin2(origin, date, result):
    calc = origin.is_valid_date(date)
    assert result == calc


def test_create_keys():
    date_start = "2001-10-10T12:00:00"
    date_end = "2002-10-10T12:00:00"
    name = "origin"
    cuuid = "c27287cd-90d1-417c-8238-f5ff220f818d"

    origin = ElementOrigin.from_keys(
        name=name, uuid=cuuid, date_start=date_start, date_end=date_end
    )
    assert isinstance(origin, ElementOrigin)
    assert origin.name == name
    assert str(origin.uuid) == cuuid
    assert origin.date_start.isoformat() == date_start
    assert origin.date_end.isoformat() == date_end
    # assert origin.description == description


def test_create_dict():
    values = dict(
        date_start="2001-10-10T12:00:00",
        date_end="2002-10-10T12:00:00",
        name="origin",
        uuid="c27287cd-90d1-417c-8238-f5ff220f818d",
    )

    origin = ElementOrigin.from_dict(values)
    assert isinstance(origin, ElementOrigin)
    assert origin.name == values["name"]
    assert str(origin.uuid) == values["uuid"]
    assert origin.date_start.isoformat() == values["date_start"]
    assert origin.date_end.isoformat() == values["date_end"]
    assert origin.description == values.get("description", "")



#
# Copyright 2019-2023 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#


import datetime
import collections.abc
import uuid

import numpy
import astropy.io.fits as fits

import numina.array.combine as C
from numina.frame.schema import SchemaKeyword as Keyword


def _m_base(calc, img, mask, region):
    full_f = img[0].data
    arr = full_f[region]
    if mask is not None:
        mask_sub = mask[region]
        invalid = (mask_sub > 0)
        arr = arr[~invalid]
    return calc(arr)


def _m_mean(img, mask, region):
    return _m_base(numpy.mean, img, mask, region)


def _m_median(img, mask, region):
    return _m_base(numpy.median, img, mask, region)


def _m_mode(img, mask, region):
    from numina.array.mode import mode_half_sample
    return _m_base(mode_half_sample, img, mask, region)


_method_map = {
    'mean': _m_mean,
    'median': _m_median,
    'mode': _m_mode
}


class Extension(object):
    def __init__(self, name):
        self.name = name


class _ExtractKeyword(object):
    def __init__(self, keyword):
        self.keyword = keyword

    def __call__(self, img, mask, region):
        return img[0].header[self.keyword.name]


def _inspect_method(value):
    if value is None:
        return value, False
    elif isinstance(value, str):
        try:
            return _method_map[value], True
        except KeyError:
            print(f'invalid method {value}')
            raise
    elif isinstance(value, collections.abc.Sequence):
        return value, False
    elif isinstance(value, Keyword):
        return _ExtractKeyword(value), True
    elif callable(value):
        return value, True
    else:
        raise TypeError('no callable')


def combine(method, images, masks=None, dtype=None,
            region=None, zeros=None, scales=None, weights=None,
            include_variance=True, datamodel=None, method_name=""):
    """Combine HDUList objects using algorithm 'method'"""

    import numina.datamodel

    nimages = len(images)

    # processing masks
    if masks is None:
        intl_masks = [None for _ in images]
    elif isinstance(masks, Extension):
        intl_masks = [img[masks.name].data for img in images]
    elif isinstance(masks, collections.abc.Sequence):
        intl_masks = [mask[0].data for mask in masks]
    else:
        raise TypeError('mask in invalid')

    if len(intl_masks) != nimages:
        raise TypeError('len(masks) != len(images)')

    # Processing region
    if region is None:
        region = Ellipsis  # meaning arr[...]

    # processing zeros, scales and weights
    num_values = dict(zeros=None, scales=None, weights=None)
    arg_values = dict(zeros=zeros, scales=scales, weights=weights)

    for key in num_values:
        arg = arg_values[key]  # locals()[key]
        value_arg, arg_is_func = _inspect_method(arg)
        if arg_is_func:
            num_values[key] = [value_arg(img, mask, region)
                               for img, mask in zip(images, intl_masks)]
        else:
            num_values[key] = value_arg
        if num_values[key] is not None:
            # Intl combine works only with np arrays
            num_values[key] = numpy.array(num_values[key])

    arrays = [hdu[0].data for hdu in images]
    if masks is None:
        fmasks = None
    else:
        fmasks = intl_masks

    out = C.generic_combine(
        method, arrays, masks=fmasks, dtype=dtype, out=None,
        zeros=num_values['zeros'], scales=num_values['scales'],
        weights=num_values['weights'])

    # Build HDUList
    headers = [img[0].header for img in images]

    base_header = headers[0].copy()
    hdu1 = fits.PrimaryHDU(data=out[0], header=base_header)
    list_of_hdu = [hdu1]

    # _logger.debug('update result header')
    prolog = ""
    if prolog:
        # _logger.debug('write prolog')
        hdu1.header['history'] = prolog

    if method_name != "":
        # _logger.info("Combined %d images using '%s'", nimages, method.__name__)
        hdu1.header['history'] = f"Combined {nimages:d} images using '{method}'"
    else:
        hdu1.header['history'] = f"Combined {nimages:d} images"

    t_str = datetime.datetime.now(datetime.timezone.utc).isoformat()
    hdu1.header['history'] = f'Combination time {t_str}'

    if datamodel is None:
        datamodel = numina.datamodel.DataModel()

    for idx, img in enumerate(images, start=1):
        hdu1.header['history'] = f"Image{idx} {datamodel.get_imgid(img)}"

    prevnum = base_header.get('NUM-NCOM', 1)

    hdu1.header['NUM-NCOM'] = prevnum * nimages
    hdu1.header['UUID'] = str(uuid.uuid1())

    # Headers of last image
    # hdu1.header['TSUTC2'] = headers[-1]['TSUTC2']

    if include_variance:
        varhdu = fits.ImageHDU(out[1], name='VARIANCE')
        list_of_hdu.append(varhdu)

    num = fits.ImageHDU(out[2].astype('uint8'), name='MAP')
    list_of_hdu.append(num)

    result = fits.HDUList(list_of_hdu)
    return result



#
# Copyright 2008-2024 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

"""User command line interface of Numina."""

import contextlib
import datetime
import logging
import os

import numina.exceptions
from numina.user.logconf import LOGCONF

from numina.util.context import working_directory
from numina.util.fqn import fully_qualified_name
from numina.user.helpers import ProcessingTask, DataManager


_logger = logging.getLogger(__name__)


def run_reduce(datastore: DataManager, obsid, as_mode=None, requirements=None, copy_files=False,
               validate_inputs=False, validate_results=False, strict_inputs=False) -> ProcessingTask:
    """Observing mode processing mode of numina."""

    request = 'reduce'
    request_params = dict()

    request_params['oblock_id'] = obsid
    request_params["pipeline"] = 'default'  # args.pipe_name
    request_params["instrument_configuration"] = 'auto'  # args.insconf
    request_params["intermediate_results"] = True
    request_params["validate_results"] = validate_results
    request_params["validate_inputs"] = validate_inputs
    request_params["strict_inputs"] = strict_inputs
    request_params["copy_files"] = copy_files
    request_params["mode"] = as_mode

    requirements = {} if requirements is None else requirements
    request_params["requirements"] = requirements

    formater_extended = LOGCONF['formatters']['extended']

    # If we put the name of logging it in LOGCONF, the file is created when
    # logging system is initialized, we don't want that
    # handler_file = LOGCONF['handlers']['processing_file']['filename']
    handler_file = "processing.log"
    logger_control = dict(
        default=__name__,
        root_levels=['numina'],
        logfile=handler_file,
        format=formater_extended['format'],
        enabled=True
    )
    request_params['logger_control'] = logger_control

    task = datastore.backend.new_task(request, request_params)

    # We should control here any possible failure
    try:
        return run_task_reduce(task, datastore)
    finally:
        datastore.store_task(task)


def run_task_reduce(task: ProcessingTask, datastore: DataManager) -> ProcessingTask:

    obsid = task.request_params['oblock_id']
    request_profile = task.request_params["instrument_configuration"]
    as_mode = task.request_params["mode"]

    _logger.info(f"procesing OB with id={obsid}")

    workenv = datastore.create_workenv(task)
    task.request_runinfo["results_dir"] = workenv.resultsdir_rel
    task.request_runinfo["work_dir"] = workenv.workdir_rel
    workenv.sane_work()

    # Roll back to cwd after leaving the context
    with working_directory(workenv.datadir):

        # here the configuration object has been updated
        # and configured with one image
        # if we have a ResultOf, ObservingMode.build_ob will insert results
        obsres = datastore.backend.obsres_from_oblock_id(
            obsid, as_mode=as_mode, configuration=request_profile
        )
        _logger.debug("instrument profile is %s", obsres.profile)
        task.request_params["instrument_configuration"] = obsres.profile
        # Merge requirements passed from above
        obsres.requirements.update(task.request_params['requirements'])
        obsres.pipeline = task.request_params["pipeline"]
        _logger.debug("pipeline is %s", obsres.pipeline)

        recipe = datastore.backend.search_recipe_from_ob(obsres)
        _logger.debug('recipe class is %s', recipe.__class__)

        recipe.intermediate_results = task.request_params["intermediate_results"]
        recipe.validate_inputs = task.request_params["validate_inputs"]
        recipe.validate_results = task.request_params["validate_results"]

        # Update runinfo
        recipe.runinfo['runner'] = task.request_runinfo['runner']
        recipe.runinfo['runner_version'] = task.request_runinfo['runner_version']
        recipe.runinfo['task_id'] = task.id
        recipe.runinfo['data_dir'] = workenv.datadir
        recipe.runinfo['work_dir'] = workenv.workdir
        recipe.runinfo['results_dir'] = workenv.resultsdir
        recipe.runinfo['intermediate_results'] = task.request_params["intermediate_results"]

        _logger.debug('recipe created')

        try:
            # here the tags have been updated
            rinput = recipe.build_recipe_input(obsres, datastore.backend)
        except (ValueError, numina.exceptions.ValidationError) as err:
            _logger.error("During recipe input construction")
            _logger.error("%s", err)
            raise

        _logger.debug('recipe input created')
        # Show the actual inputs
        for key, val in obsres.requirements.items():
            if key not in recipe.requirements():
                msg = f'"{key}: {val}" present in OB requirements, but not used'
                if task.request_params['strict_inputs']:
                    _logger.error(msg)
                    raise ValueError(msg)
                else:
                    _logger.warning(msg)

        for key, req in recipe.requirements().items():
            v = getattr(rinput, key)
            _logger.debug("recipe requires %r, value is %s", key, v)

        for req in recipe.products().values():
            _logger.debug('recipe provides %s, %s',
                          req.type.__class__.__name__, req.description)

    # Load recipe control and recipe parameters from file
    task.request_runinfo['instrument'] = obsres.instrument
    task.request_runinfo['pipeline'] = obsres.pipeline
    task.request_runinfo['mode'] = obsres.mode
    task.request_runinfo['recipe_class'] = recipe.__class__.__name__
    task.request_runinfo['recipe_fqn'] = fully_qualified_name(recipe.__class__)
    task.request_runinfo['recipe_version'] = recipe.__version__
    root_level_logger = task.request_runinfo['recipe_fqn'].split('.')[0]

    # Copy files
    if task.request_params["copy_files"]:
        install_action = 'copy'
        _logger.debug('copy files to work directory')
    else:
        install_action = 'link'
        _logger.debug('link files to work directory')

    workenv.installfiles_stage1(obsres, action=install_action)
    workenv.installfiles_stage2(rinput, action=install_action)
    workenv.adapt_obsres(obsres)

    logger_control = task.request_params['logger_control']
    # configure recipe root level
    with config_recipe_logger(root_level_logger):
        logger_control['root_levels'].append(root_level_logger)
        # Add file logging
        with logger_manager(logger_control, workenv.resultsdir):
            with working_directory(workenv.workdir):
                completed_task = run_recipe_timed(task, recipe, rinput)

    # datastore.store_task(completed_task)
    return completed_task


def run_recipe_timed(task: ProcessingTask, recipe, rinput) -> ProcessingTask:
    """Run the recipe and count the time it takes."""
    _logger.info('running recipe')
    task.state = 1
    task.time_start = datetime.datetime.now()
    #
    try:
        task.result = recipe(rinput)
        task.state = 2
    except Exception:
        task.state = 3
        raise
    finally:
        task.time_end = datetime.datetime.now()
    return task


@contextlib.contextmanager
def logger_manager(logger_control, result_dir):
    """"Add a FileHandler to existing loggers

    We add a FileHandler to the loggers named in
    logger_control['root_levels'] if
    logger_control['root_levels'] is True
    with level DEBUG

    After exiting the context, the handlers are removed
    """
    # Creating custom logger file
    if logger_control['root_levels']:
        logfile = os.path.join(result_dir, logger_control['logfile'])
        logformat = logger_control['format']
        recipe_formatter = logging.Formatter(logformat)
        fh = logging.FileHandler(logfile, mode='w')
        fh.setLevel(logging.DEBUG)
        fh.setFormatter(recipe_formatter)
    else:
        fh = logging.NullHandler()

    root_levels = logger_control['root_levels']
    recipe_loggers = []
    for lname in root_levels:
        recipe_logger = logging.getLogger(lname)
        recipe_logger.setLevel(logging.DEBUG)
        recipe_logger.addHandler(fh)
        recipe_loggers.append(recipe_logger)
    try:
        yield recipe_loggers
    finally:
        for recipe_logger in recipe_loggers:
            recipe_logger.removeHandler(fh)


@contextlib.contextmanager
def config_recipe_logger(root_level_logger, ref_logger='numina'):
    """Configure root level logger for recipe.

    The recipe formatter is 'detailed'
    The logger is set to DEBUG (this is needed by the FileHandler later
    The StreamHandler is set to the same level as the logger of numina

    """
    numina_logger = logging.getLogger(ref_logger)
    recipe_logger = logging.getLogger(root_level_logger)
    # Save state of logger
    save_level = recipe_logger.getEffectiveLevel()
    recipe_logger.setLevel(logging.DEBUG)
    save_propagate = recipe_logger.propagate
    recipe_logger.propagate = False

    # create formatter
    formater_dd = LOGCONF['formatters']['detailed']
    detailed_formatter = logging.Formatter(fmt=formater_dd.get('format'))
    # create handler, ignoring configuration here
    # handerl_dd = numina_cli_logconf['handlers']['detailed_console']
    # handerl_dd_level =
    sh = logging.StreamHandler()
    sh.setLevel(numina_logger.getEffectiveLevel())
    sh.setFormatter(detailed_formatter)
    recipe_logger.addHandler(sh)

    try:
        yield recipe_logger
    finally:
        # Restore state
        recipe_logger.removeHandler(sh)
        recipe_logger.propagate = save_propagate
        recipe_logger.setLevel(save_level)



#
# Copyright 2025 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

"""Generate a 2D mosaic from individual 2D images

The images to be combined are resampled to the same WCS using the
reproject package. Masks can be used in the combination process.
The masks are the ones available in some extension of the individual
images or can be calculated in real-time from the values of the
images containing np.nan.

All images resampled to the same WCS are stored in a masked array,
allowing the combination to be performed using different functions:
mean, median, sum, std, sigmaclip_mean, sigmaclip_median, and
sigmaclip_stddev.
"""

import argparse
from astropy.io import fits
from astropy.stats import sigma_clipped_stats
from astropy.wcs import WCS
import numpy as np
import numpy.ma as ma
from reproject import reproject_interp, reproject_adaptive, reproject_exact
from reproject.mosaicking import find_optimal_celestial_wcs
import sys

from numina.instrument.simulation.ifu.define_3d_wcs import wcs_to_header_using_cd_keywords

from .ctext import ctext

REPROJECT_METHODS = ['interp', 'adaptive', 'exact']
COMBINATION_FUNCTIONS = ['mean', 'median', 'sum', 'std','sigmaclip_mean', 'sigmaclip_median', 'sigmaclip_stddev']


def generate_mosaic_of_2d_images(
        list_of_hdu2d_images,
        list_of_hdu2d_masks=None,
        reproject_method=None,
        combination_function=None,
        verbose=False
):
    """Combine 2D images using their WCS information.

    Parameters
    ----------
    list_of_hdu2d_images : list of HDU images
        List of 2D HDU instances containing the images to be combined.
    list_of_hdu2d_masks : list or None
        List of 2D HDU instances containing the masks associated to
        the images to be combined. If this list is None, this function
        computes a particular mask for each 2D image prior to the
        combination, looking for np.nan values.
    reproject_method : str
        Reprojection method. See 'REPROJECT_METHODS' above.
    combination_function : str
        Combination function. See 'COMBINATION_FUNCTIONS' above.
    verbose : bool
        If True, display additional information.

    Returns
    -------
    mosaic : numpy.ma.MaskedArray
        Masked array containing the combined image and the resulting
        mask.
    wcs_mosaic2d: astropy.wcs.WCS
        Instance representing the World Coordinate System of the
        combined image.
    stack3d : numpy.ma.MaskedArray
        Stacked masked array containing the individual reprojected
        images and masks.
    """

    # protections
    if not isinstance(list_of_hdu2d_images, list):
        raise TypeError('list_of_hdu2d_images must be a list')
    if not isinstance(list_of_hdu2d_masks, list) and list_of_hdu2d_masks is not None:
        raise TypeError('list_of_hdu2d_masks must be a list or None')

    nimages = len(list_of_hdu2d_images)
    if verbose:
        print(f'Total number of images to be combined: {nimages}')

    if nimages < 1:
        raise ValueError('Number of images = 0')

    # check masks and generate them if not present
    if list_of_hdu2d_masks is None:
        list_of_hdu2d_masks = []
        for hdu2d_image in list_of_hdu2d_images:
            hdu2d_mask = fits.ImageHDU(data=np.isnan(hdu2d_image.data).astype(np.uint8))
            hdu2d_mask.header.extend(wcs_to_header_using_cd_keywords(WCS(hdu2d_image.header)), update=True)
            list_of_hdu2d_masks.append(hdu2d_mask)
    else:
        if len(list_of_hdu2d_images) != len(list_of_hdu2d_masks):
            raise ValueError(f'Unexpected {len(list_of_hdu2d_images)=} != {len(list_of_hdu2d_masks)=}')

    # check compatibility between image and mask dimensions
    for hdu2d_image, hdu2d_mask in zip(list_of_hdu2d_images, list_of_hdu2d_masks):
        if hdu2d_image.shape != hdu2d_mask.shape:
            raise ValueError(f'Unexpected shape {hdu2d_image.shape=} != {hdu2d_mask.shape=}')

    # compute optimal WCS for combined image
    wcs_mosaic2d, shape_mosaic2d = find_optimal_celestial_wcs(list_of_hdu2d_images)
    if verbose:
        print(f'\n{wcs_mosaic2d=}')
        print(f'{shape_mosaic2d=}')
    naxis2_mosaic, naxis1_mosaic = shape_mosaic2d

    # we are transforming the individual images and storing the result
    # in an auxiliary 3D masked cube
    shape3d = nimages, naxis2_mosaic, naxis1_mosaic
    stack3d = ma.array(np.zeros(shape3d), mask=np.full(shape3d, fill_value=False))

    # generate mosaic
    if verbose:
        print(f'Reprojection method: {reproject_method}')
    for i in range(nimages):
        hdu2d_image = list_of_hdu2d_images[i]
        hdu2d_mask = list_of_hdu2d_masks[i]
        if reproject_method == 'interp':
            image2d_resampled, footprint_image2d_resampled = reproject_interp(
                input_data=hdu2d_image,
                output_projection=wcs_mosaic2d,
                shape_out=shape_mosaic2d
            )
            if hdu2d_mask is not None:
                mask2d_resampled, _ = reproject_interp(
                    input_data=hdu2d_mask,
                    output_projection=wcs_mosaic2d,
                    shape_out=shape_mosaic2d
                )
            else:
                mask2d_resampled = None
        elif reproject_method == 'adaptive':
            image2d_resampled, footprint_image2d_resampled = reproject_adaptive(
                input_data=hdu2d_image,
                output_projection=wcs_mosaic2d,
                shape_out=shape_mosaic2d,
                conserve_flux=True,
                kernel='Gaussian'
            )
            if hdu2d_mask is not None:
                mask2d_resampled, _ = reproject_adaptive(
                    input_data=hdu2d_mask,
                    output_projection=wcs_mosaic2d,
                    shape_out=shape_mosaic2d,
                    conserve_flux=True,
                    kernel='Gaussian'
                )
            else:
                mask2d_resampled = None
        elif reproject_method == 'exact':
            image2d_resampled, footprint_image2d_resampled = reproject_exact(
                input_data=hdu2d_image,
                output_projection=wcs_mosaic2d,
                shape_out=shape_mosaic2d,
            )
            if hdu2d_mask is not None:
                mask2d_resampled, _ = reproject_exact(
                    input_data=hdu2d_mask,
                    output_projection=wcs_mosaic2d,
                    shape_out=shape_mosaic2d,
                )
            else:
                mask2d_resampled = None
        else:
            raise ValueError(f'Unexpected {reproject_method=}')
        stack3d[i, :, :] = image2d_resampled
        if mask2d_resampled is None:
            mask2d_resampled = np.full(shape_mosaic2d, fill_value=False)
            stack3d[i, :, :].mask = mask2d_resampled
        else:
            # merge footprints
            stack3d[i, :, :].mask = np.logical_or(
                np.logical_or(mask2d_resampled, mask2d_resampled==np.nan),
                (footprint_image2d_resampled == 0)
            )

    # generate mosaic
    if combination_function == 'mean':
        mosaic = stack3d.mean(axis=0)
    elif combination_function == 'median':
        mosaic = stack3d.median(axis=0)
    elif combination_function == 'sum':
        mosaic = stack3d.sum(axis=0)
    elif combination_function == 'std':
        mosaic = stack3d.std(axis=0)
    elif combination_function[:9] == 'sigmaclip':
        array2d_sigclip_mean, array2d_sigclip_median, array2d_sigclip_stddev = \
            sigma_clipped_stats(stack3d, axis=0, maxiters=None)
        if combination_function == 'sigmaclip_mean':
            mosaic = array2d_sigclip_mean
        elif combination_function == 'sigmaclip_median':
            mosaic = array2d_sigclip_median
        elif combination_function == 'sigmaclip_stddev':
            mosaic = array2d_sigclip_stddev
        else:
            raise ValueError(f'Unexpected {combination_function=}')
    else:
        raise ValueError(f'Unexpected {combination_function=}')

    if verbose:
        print(f'Combination function: {combination_function}')

    # set to np.nan the masked values
    mosaic.data[mosaic.mask] = np.nan

    # return result
    return mosaic, wcs_mosaic2d, stack3d


def main(args=None):

    # parse command-line options
    parser = argparse.ArgumentParser()
    parser.add_argument("input_list",
                        help="TXT file with list of 2D images to be combined", type=str)
    parser.add_argument('output_filename',
                        help='filename of output FITS image', type=str)
    parser.add_argument('--reproject_method',
                        help='Reprojection method (interp, adaptive, exact)',
                        type=str, choices=REPROJECT_METHODS, default='adaptive')
    parser.add_argument('--extname_image',
                        help='Extension name for image in input files. Default value: PRIMARY',
                        default='PRIMARY', type=str)
    parser.add_argument('--extname_mask',
                        help="Extension name for mask in input files. Default 'None': use np.nan in image",
                        default=None, type=str)
    parser.add_argument("--combination_function", help='Combination function. Default: mean',
                        type=str, default='mean', choices=COMBINATION_FUNCTIONS)
    parser.add_argument("--output_3D_stack",
                        help="filename for stacked 3D array. Default None",
                        default=None, type=str)
    parser.add_argument("--verbose",
                        help="Display intermediate information",
                        action="store_true")
    parser.add_argument("--echo",
                        help="Display full command line",
                        action="store_true")

    args = parser.parse_args(args)

    if len(sys.argv) == 1:
        parser.print_usage()
        raise SystemExit()

    if args.verbose:
        for arg, value in vars(args).items():
            print(ctext(f'{arg}: {value}', faint=True))

    if args.echo:
        print('\033[1m\033[31mExecuting: ' + ' '.join(sys.argv) + '\033[0m\n')

    input_list = args.input_list
    output_filename = args.output_filename
    reproject_method = args.reproject_method
    combination_function = args.combination_function
    output_3d_stack = args.output_3D_stack
    verbose = args.verbose

    # define extensions for image and mask
    extname_image = args.extname_image
    extname_mask = args.extname_mask
    if extname_mask is not None:
        if extname_mask.lower() == 'none':
            extname_mask = None

    # read input file with list of 2D images to be combined
    with open(input_list) as f:
        file_content = f.read().splitlines()

    # list of HDU
    list_of_hdu2d_images = []
    list_of_hdu2d_masks = []
    for fname in file_content:
        if len(fname) > 0:
            if fname[0] not in ['#']:
                print(f'\n* Reading: {fname}')
                with fits.open(fname) as hdul:
                    if extname_image not in hdul:
                        raise ValueError(f'Expected {extname_image} extension not found')
                    hdu2d_image = hdul[extname_image].copy()
                    if verbose:
                        print(f"{hdu2d_image.header['NAXIS1']=}")
                        print(f"{hdu2d_image.header['NAXIS2']=}")
                    if extname_mask is not None:
                        if extname_mask in hdul:
                            hdu2d_mask = hdul[extname_mask].copy()
                            if verbose:
                                print(f'image mask: {extname_mask}')
                            bitpix_mask = hdu2d_mask.header['BITPIX']
                            if bitpix_mask != 8:
                                raise ValueError(f'BITPIX (mask): {bitpix_mask} is not 8')
                            if hdu2d_image.data.shape != hdu2d_mask.data.shape:
                                raise ValueError(f'Shape of {extname_image} and {extname_mask} are different!')
                        else:
                            hdu2d_mask = None
                    else:
                        hdu2d_mask = None
                    # generate mask from np.nan when necessary
                    if hdu2d_mask is None:
                        hdu2d_mask = fits.ImageHDU(data=np.isnan(hdu2d_image.data).astype(np.uint8))
                        hdu2d_mask.header.extend(wcs_to_header_using_cd_keywords(WCS(hdu2d_image.header)), update=True)
                        if verbose:
                            print(f'generating mask from np.nan values')
                    if verbose:
                        print(f'Number of masked pixels / total: {np.sum(hdu2d_mask.data)} / {hdu2d_mask.size}')
                    # store image and associated footprint
                    list_of_hdu2d_images.append(hdu2d_image)
                    list_of_hdu2d_masks.append(hdu2d_mask)

    # combine images
    mosaic, wcs_mosaic2d, stack3d = generate_mosaic_of_2d_images(
        list_of_hdu2d_images=list_of_hdu2d_images,
        list_of_hdu2d_masks=list_of_hdu2d_masks,
        reproject_method=reproject_method,
        combination_function=combination_function,
        verbose=verbose
    )

    # save result
    hdu = fits.PrimaryHDU(mosaic.data.astype(np.float32))
    hdu.header.extend(wcs_to_header_using_cd_keywords(wcs_mosaic2d), update=True)
    hdu_mask = fits.ImageHDU(data=mosaic.mask.astype(np.uint8))
    hdu_mask.header['EXTNAME'] = 'MASK'
    hdu_mask.header.extend(wcs_to_header_using_cd_keywords(wcs_mosaic2d), update=True)
    hdul = fits.HDUList([hdu, hdu_mask])
    if verbose:
        print(f'\nSaving combined 2D image: {output_filename}')
    hdul.writeto(output_filename, overwrite='yes')

    # save 3D stack if requested
    if output_3d_stack is not None:
        hdu = fits.PrimaryHDU(stack3d.data.astype(np.float32))
        hdu.header.extend(wcs_to_header_using_cd_keywords(wcs_mosaic2d), update=True)
        hdu_mask = fits.ImageHDU(data=stack3d.mask.astype(np.uint8))
        hdu_mask.header['EXTNAME'] = 'MASK'
        hdu_mask.header.extend(wcs_to_header_using_cd_keywords(wcs_mosaic2d), update=True)
        hdul = fits.HDUList([hdu, hdu_mask])
        if verbose:
            print(f'Saving 3D stack.........: {output_3d_stack}')
        hdul.writeto(output_3d_stack, overwrite='yes')


if __name__ == "__main__":

    main()



#
# Copyright 2016-2018 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

import numpy


class Efficiency(object):

    def response(self, wl):
        return numpy.ones_like(wl)



#
# Copyright 2025 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#
"""Computation of cosmic ray masks using several equivalent exposures."""
import ast
import inspect
import io
import logging
import sys
import uuid

from astropy.io import fits
from ccdproc import cosmicray_lacosmic
from contextlib import redirect_stderr, redirect_stdout
from datetime import datetime
from matplotlib.colors import LogNorm
from matplotlib.colors import LinearSegmentedColormap
import matplotlib.pyplot as plt
import numpy as np
from rich.logging import RichHandler
from scipy import ndimage
from skimage.registration import phase_cross_correlation

from numina.array.distortion import shift_image2d
from numina.array.numsplines import spline_positive_derivative
import teareduce as tea

from .valid_parameters import VALID_CRMETHODS
from .valid_parameters import VALID_LACOSMIC_CLEANTYPE
from .valid_parameters import VALID_BOUNDARY_FITS
from .all_valid_numbers import all_valid_numbers
from .compute_flux_factor import compute_flux_factor
from .define_piecewise_linear_function import define_piecewise_linear_function
from .diagnostic_plot import diagnostic_plot
from .display_detected_cr import display_detected_cr
from .estimate_diagnostic_limits import estimate_diagnostic_limits
from .gausskernel2d_elliptical import gausskernel2d_elliptical


def decorate_output(func):
    """Decorator to capture stdout and stderr of a function and log it."""

    def wrapper(*args, **kwargs):
        _logger = logging.getLogger(__name__)
        buf = io.StringIO()
        with redirect_stdout(buf), redirect_stderr(buf):
            result = func(*args, **kwargs)
        # Split into lines
        output_lines = buf.getvalue().splitlines()
        # Remove trailing empty line
        if output_lines and output_lines[-1] == '':
            output_lines = output_lines[:-1]
        if output_lines:
            _logger.info("\n".join(output_lines))
        return result

    return wrapper


@decorate_output
def decorated_cosmicray_lacosmic(*args, **kwargs):
    """Wrapper for cosmicray_lacosmic with decorated output."""
    return cosmicray_lacosmic(*args, **kwargs)


def compute_crmasks(
        list_arrays,
        gain=None,
        rnoise=None,
        bias=None,
        crmethod='mm_lacosmic',
        use_lamedian=False,
        flux_factor=None,
        flux_factor_regions=None,
        apply_flux_factor_to=None,
        interactive=True,
        dilation=1,
        regions_to_be_skipped=None,
        pixels_to_be_flagged_as_cr=None,
        pixels_to_be_ignored_as_cr=None,
        pixels_to_be_replaced_by_local_median=None,
        dtype=np.float32,
        verify_cr=False,
        semiwindow=15,
        color_scale='minmax',
        maxplots=-1,
        debug=False,
        _logger=None,
        la_gain_apply=True,
        la_sigclip=None,
        la_sigfrac=None,
        la_objlim=None,
        la_satlevel=None,
        la_niter=None,
        la_sepmed=None,
        la_cleantype=None,
        la_fsmode=None,
        la_psfmodel=None,
        la_psffwhm_x=None,
        la_psffwhm_y=None,
        la_psfsize=None,
        la_psfbeta=None,
        la_verbose=False,
        mm_xy_offsets=None,
        mm_crosscorr_region=None,
        mm_boundary_fit=None,
        mm_knots_splfit=3,
        mm_fixed_points_in_boundary=None,
        mm_nsimulations=10,
        mm_niter_boundary_extension=3,
        mm_weight_boundary_extension=10.0,
        mm_threshold=0.0,
        mm_minimum_max2d_rnoise=5.0,
        mm_seed=None
        ):
    """
    Computation of cosmic rays masks using several equivalent exposures.

    This function computes cosmic ray masks from a list of 2D numpy arrays.
    Two different methods are implemented:
    1. Cosmic ray detection using the Laplacian edge detection algorithm
       (van Dokkum 2001), as implemented in ccdproc.cosmicray_lacosmic.
    2. Cosmic ray detection using a numerically derived boundary in the
       median combined image. The cosmic ray detection is based on a boundary
       that is derived numerically making use of the provided gain and readout
       noise values. The function also supports generating diagnostic plots to
       visualize the cosmic ray detection process.

    Parameters
    ----------
    list_arrays : list of 2D arrays
        The input arrays to be combined. The arrays are assumed to be
        provided in ADU.
    gain : 2D array, float or None
        The gain value (in e/ADU) of the detector.
        If None, it is assumed to be 1.0.
    rnoise : 2D array, float or None
        The readout noise (in ADU) of the detector.
        If None, it is assumed to be 0.0.
    bias : 2D array, float or None
        The bias value (in ADU) of the detector.
        If None, it is assumed to be 0.0.
    crmethod : str
        The method to use for cosmic ray detection. Valid options are:
        - 'lacosmic': use the cosmic-ray rejection by Laplacian edge
        detection (van Dokkum 2001), as implemented in ccdproc.
        - 'mmcosmic': use the numerically derived boundary to
        detect cosmic rays in the median combined image.
        - 'mm_lacosmic': use both methods: 'lacosmic' and 'mmcosmic'.
        Pixels detected by either method are included in the final mask.
    use_lamedian: bool, optional
        If True, use the corrected values from the lacosmic algorithm
        when replacing the cosmic-ray affected pixels in the median
        combined image. This parameter is only used when
        crmethod is 'lacosmic', or 'mm_lacosmic'.
    flux_factor : str, list, float or None, optional
        The flux scaling factor for each exposure (default is None).
        If 'auto', the flux factor is determined automatically.
        If None or 'none', it is set to 1.0 for all images.
        If a float is provided, it is used as the flux factor for all images.
        If a list is provided, it should contain a value
        for each single image in `list_arrays`.
    flux_factor_regions : list of lists of 4 integers or None, optional
        The regions to use for computing the flux scaling factors
        when flux_factor='auto'. If None, the entire image is used.
        The format of the each region must be a list of 4 integers, following
        the FITS convention [xmin, xmax, ymin, ymax], where the indices
        start from 1 to NAXIS[12].
    apply_flux_factor_to : str
        Specifies to which images the flux factor should be applied.
        Valid options are:
        - 'original': apply the flux factor to the original images.
        - 'simulated': apply the flux factor to the simulated images
          used to derive the boundary.
    interactive : bool, optional
        If True, enable interactive mode for plots.
    dilation : int, optional
        The dilation factor for the coincident cosmic-ray pixel mask.
    regions_to_be_skipped : list of lists of 4 integers or None, optional
        The regions to be skipped during cosmic ray detection.
        If None, no regions are skipped. The format of each region must
        be a list of 4 integers, following the FITS convention
        [xmin, xmax, ymin, ymax], where the indices start from 1 to NAXIS[12].
    pixels_to_be_flagged_as_cr : list of (x, y) tuples, or None, optional
        List of pixel coordinates to be included in the masks
        (Assuming FITS criterium; first pixel is (1, 1)).
    pixels_to_be_ignored_as_cr : list of (x, y) tuples, or None, optional
        List of pixel coordinates to be excluded from the masks
        (Assuming FITS criterium; first pixel is (1, 1)).
    pixels_to_be_replaced_by_local_median : list of (x, y, x_with, y_width) tuples, or None, optional
        List of pixel coordinates to be replaced by the median value
        when removing the cosmic rays. This information is stored as a binary
        table in one of the extensions of the output HDUList with the masks.
        (Assuming FITS criterium; first pixel is (1, 1)).
    dtype : data-type, optional
        The desired data type to build the 3D stack (default is np.float32).
    verify_cr : bool, optional
        If True, verify the cosmic ray detection by comparing the
        detected positions with the original images (default is True). It
        is only used if interactive=True.
    semiwindow : int, optional
        The semiwindow size to plot the coincident cosmic-ray pixels (default is 15).
    color_scale : str, optional
        The color scale to use for the plots (default is 'minmax').
        Valid options are 'minmax' and 'zscale'.
    maxplots : int, optional
        The maximum number of coincident cosmic-ray pixels to plot (default is -1).
        If negative, all detected cosmic-ray pixels will be plotted.
    debug : bool, optional
        If True, enable debug mode (default is False).
    _logger : logging.Logger or None, optional
        The logger to use for logging. If None, a new logger is created.
    la_gain_apply: bool, optional
        If True, apply the gain when computing the cosmic ray mask
        with the lacosmic algorithm. Default is True.
    la_sigclip : float
        The sigma clipping threshold. Employed when crmethod='lacosmic'.
    la_sigfrac : float
        The fractional detection limit for neighboring pixels.
        Employed when crmethod='lacosmic'.
    la_objlim : float
        Minimum contrast between Laplacian image and fine structure image.
        Employed when crmethod='lacosmic'.
    la_satlevel : float
        The saturation level (in ADU) of the detector. Employed when crmethod='lacosmic'.
    la_niter : int
        The number of iterations to perform. Employed when crmethod='lacosmic'.
    la_sepmed : bool
        If True, use separable median filter instead of the full median filter.
        Employed when crmethod='lacosmic'.
    la_fsmode : str
        The mode to use for the fine structure image. Valid options are:
        'median' or 'convolve'. Employed when crmethod='lacosmic'.
    la_psfmodel : str
        The model to use for the PSF if la_fsmode='convolve'.
        Valid options are:
        - circular kernels: 'gauss' or 'moffat'
        - Gaussian in the x and y directions: 'gaussx' and 'gaussy'
        - elliptical Gaussian: 'gaussxy' (this kernel is not available
          in ccdproc.cosmicray_lacosmic, so it is implemented here)
        Employed when crmethod='lacosmic'.
    la_psffwhm_x : float
        The full width at half maximum (FWHM, in pixels) of the PSF in
        the x direction. Employed when crmethod='lacosmic'.
    la_psffwhm_y : float
        The full width at half maximum (FWHM, in pixels) of the PSF
        in the y direction. Employed when crmethod='lacosmic'.
    la_psfsize : int
        The kernel size to use for the PSF. It must be an odd integer >= 3.
        Employed when crmethod='lacosmic'.
    la_psfbeta : float
        The beta parameter of the Moffat PSF. It is only used if
        la_psfmodel='moffat'. Employed when crmethod='lacosmic'.
    la_verbose : bool
        If True, print additional information during the
        execution. Employed when crmethod='lacosmic'.
    mm_xy_offsets: list of [x_offset, y_offset] or None
        The offsets to apply to each simulated individual exposure
        when computing the diagnostic diagram for the mmcosmic method.
        If None, no offsets are applied.
        This option is not compatible with 'mm_crosscorr_region'.
    mm_crosscorr_region : list of 4 integers or None
        The region to use for the 2D cross-correlation to determine
        the offsets between the individual images and the median image.
        If None, no offsets are computed and it is assumed that
        the images are already aligned. The format of the region
        must follow the FITS convention [xmin, xmax, ymin, ymax],
        where the indices start from 1 to NAXIS[12].
        This option is not compatible with 'mm_xy_offsets'.
    mm_boundary_fit : str, or None
        The method to use for the boundary fitting. Valid options are:
        - 'spline': use a spline fit to the boundary.
        - 'piecewise': use a piecewise linear fit to the boundary.
    mm_knots_splfit : int, optional
        The number of knots for the spline fit to the boundary.
    mm_fixed_points_in_boundary : str, or list or None
        The fixed points to use for the boundary fitting.
    mm_nsimulations : int, optional
        The number of simulations of each set of input images to compute
        the detection boundary.
    mm_niter_boundary_extension : int, optional
        The number of iterations for the boundary extension.
    mm_weight_boundary_extension : float, optional
        The weight for the boundary extension.
        In each iteration, the boundary is extended by applying an
        extra weight to the points above the previous boundary. This
        extra weight is computed as `mm_weight_boundary_extension**iter`,
        where `iter` is the current iteration number (starting from 1).
    mm_threshold: float, optional
        Minimum threshold for median2d - min2d to consider a pixel as a
        cosmic ray (default is None). If None, the threshold is computed
        automatically from the minimum boundary value in the numerical
        simulations.
    mm_minimum_max2d_rnoise : float, optional
        Minimum value for max2d in readout noise units to flag a pixel
        as a coincident cosmic-ray pixel.
    mm_seed : int or None, optional
        The random seed for reproducibility.

    Returns
    -------
    hdul_masks : hdulist
        The HDUList containing the mask arrays for cosmic ray
        removal using different methods. The primary HDU only contains
        information about the parameters used to determine the
        suspected pixels. The extensions are:
        - 'MEDIANCR': Mask for coincident cosmic-ray pixels detected using the
        median combination.
        - 'MEANCRT': Mask for cosmic-ray pixels detected when adding all the
        individual arrays. That summed image contains all the cosmic-ray pixels.
        of all the images.
        - 'CRMASK1', 'CRMASK2', ...: Masks for cosmic-ray pixels detected
        in each individual array.
    """

    if _logger is None:
        _logger = logging.getLogger(__name__)
        rich_configured = False
    else:
        # use the provided logger
        root_logger = logging.getLogger()
        # check if RichHandler is configured in the parent logger's handler
        rich_configured = any(isinstance(handler, RichHandler) for handler in root_logger.handlers)
    # Set up rich labels for logging
    if rich_configured:
        rlabel_crmethod = f"[bold green]{crmethod}[/bold green]"
        rlabel_lacosmic = "[bold red]lacosmic[/bold red]"
        rlabel_mmcosmic = "[bold blue]mmcosmic[/bold blue]"
    else:
        rlabel_crmethod = f"{crmethod}"
        rlabel_lacosmic = "lacosmic"
        rlabel_mmcosmic = "mmcosmic"

    # Check that the input is a list
    if not isinstance(list_arrays, list):
        raise TypeError("Input must be a list of arrays.")

    # Check that the list contains numpy 2D arrays
    if not all(isinstance(array, np.ndarray) and array.ndim == 2 for array in list_arrays):
        raise ValueError("All elements in the list must be 2D numpy arrays.")

    # Check that the list contains at least 3 arrays
    num_images = len(list_arrays)
    if num_images < 3:
        raise ValueError("At least 3 images are required for useful mediancr combination.")

    # Check that all arrays have the same shape
    for i, array in enumerate(list_arrays):
        if array.shape != list_arrays[0].shape:
            raise ValueError(f"Array {i} has a different shape than the first array.")
    naxis2, naxis1 = list_arrays[0].shape

    # Log the number of input arrays and their shapes
    _logger.info("number of input arrays: %d", len(list_arrays))
    for i, array in enumerate(list_arrays):
        _logger.debug("array %d shape: %s, dtype: %s", i, array.shape, array.dtype)

    # Check that interactive is True if verify_cr is True
    if verify_cr and not interactive:
        raise ValueError("interactive must be True if verify_cr is True.")

    # Define the gain
    gain_scalar = None  # store gain as a scalar value (if applicable)
    if gain is None:
        gain = np.ones((naxis2, naxis1), dtype=float)
        _logger.warning("gain not defined, assuming gain=1.0 for all pixels.")
        gain_scalar = 1.0
    elif isinstance(gain, (float, int)):
        gain = np.full((naxis2, naxis1), gain, dtype=float)
        _logger.info("gain defined as a constant value: %f", gain[0, 0])
        gain_scalar = float(gain[0, 0])
    elif isinstance(gain, np.ndarray):
        if gain.shape != (naxis2, naxis1):
            raise ValueError(f"gain must have the same shape as the input arrays ({naxis2=}, {naxis1=}).")
        _logger.info("gain defined as a 2D array with shape: %s", gain.shape)
        if np.all(gain == gain[0, 0]):
            gain_scalar = float(gain[0, 0])
    else:
        raise TypeError(f"Invalid type for gain: {type(gain)}. Must be float, int, or numpy array.")

    # Define the readout noise
    rnoise_scalar = None  # store rnoise as a scalar value (if applicable)
    if rnoise is None:
        rnoise = np.zeros((naxis2, naxis1), dtype=float)
        _logger.warning("readout noise not defined, assuming readout noise=0.0 for all pixels.")
        rnoise_scalar = 0.0
    elif isinstance(rnoise, (float, int)):
        rnoise = np.full((naxis2, naxis1), rnoise, dtype=float)
        _logger.info("readout noise defined as a constant value: %f", rnoise[0, 0])
        rnoise_scalar = float(rnoise[0, 0])
    elif isinstance(rnoise, np.ndarray):
        if rnoise.shape != (naxis2, naxis1):
            raise ValueError(f"rnoise must have the same shape as the input arrays ({naxis2=}, {naxis1=}).")
        _logger.info("readout noise defined as a 2D array with shape: %s", rnoise.shape)
        if np.all(rnoise == rnoise[0, 0]):
            rnoise_scalar = float(rnoise[0, 0])
    else:
        raise TypeError(f"Invalid type for rnoise: {type(rnoise)}. Must be float, int, or numpy array.")

    # Define the bias
    if bias is None:
        bias = np.zeros((naxis2, naxis1), dtype=float)
        _logger.warning("bias not defined, assuming bias=0.0 for all pixels.")
    elif isinstance(bias, (float, int)):
        bias = np.full((naxis2, naxis1), bias, dtype=float)
        _logger.info("bias defined as a constant value: %f", bias[0, 0])
    elif isinstance(bias, np.ndarray):
        if bias.shape != (naxis2, naxis1):
            raise ValueError(f"bias must have the same shape as the input arrays ({naxis2=}, {naxis1=}).")
        _logger.info("bias defined as a 2D array with shape: %s", bias.shape)
    else:
        raise TypeError(f"Invalid type for bias: {type(bias)}. Must be float, int, or numpy array.")

    # Convert the list of arrays to a 3D numpy array
    image3d = np.zeros((num_images, naxis2, naxis1), dtype=dtype)
    for i, array in enumerate(list_arrays):
        image3d[i] = array.astype(dtype)

    # Subtract the bias from the input arrays
    _logger.info("subtracting bias from the input arrays")
    image3d -= bias

    # Check crmethod
    if crmethod not in VALID_CRMETHODS:
        raise ValueError(f"Invalid crmethod: {crmethod}. Valid options are {VALID_CRMETHODS}.")
    _logger.info("computing crmasks using crmethod: %s", rlabel_crmethod)
    # Check use_lamedian
    if use_lamedian and crmethod not in ['lacosmic', 'mm_lacosmic']:
        raise ValueError("use_lamedian can only be True when crmethod is 'lacosmic' or 'mm_lacosmic'.")
    _logger.info("use_lamedian: %s", str(use_lamedian))

    # Check flux_factor
    if isinstance(flux_factor_regions, str):
        if flux_factor_regions.lower() == 'none':
            flux_factor_regions = None
    if flux_factor is None:
        flux_factor = np.ones(num_images, dtype=float)
    elif isinstance(flux_factor, str):
        if flux_factor.lower() == 'auto':
            _logger.info("flux_factor set to 'auto', computing values...")
            list_flux_factor_regions = []
            if isinstance(flux_factor_regions, str):
                flux_factor_regions = ast.literal_eval(flux_factor_regions)
            if isinstance(flux_factor_regions, list):
                for flux_factor_region in flux_factor_regions:
                    if isinstance(flux_factor_region, list):
                        all_integers = all(isinstance(val, int) for val in flux_factor_region)
                        if not all_integers:
                            raise TypeError(f"Invalid flux_factor_region: {flux_factor_region}. "
                                            "All elements must be integers.")
                        if len(flux_factor_region) != 4:
                            raise ValueError(f"Invalid length for flux_factor_region: {flux_factor_region}. "
                                             "Must be a list of 4 integers [xmin, xmax, ymin, ymax].")
                        dumreg = f"[{flux_factor_region[0]}:{flux_factor_region[1]}, " + \
                                 f"{flux_factor_region[2]}:{flux_factor_region[3]}]"
                        _logger.debug("defined flux factor region: %s", dumreg)
                        ff_region = tea.SliceRegion2D(dumreg, mode='fits', naxis1=naxis1, naxis2=naxis2)
                    else:
                        raise TypeError(f"Invalid type for flux_factor_region in the list: {type(flux_factor_region)}. "
                                        "Must be a list of 4 integers")
                    list_flux_factor_regions.append(ff_region)
            elif flux_factor_regions is None:
                ff_region = tea.SliceRegion2D(f'[1:{naxis1}, 1:{naxis2}]', mode='fits')
                list_flux_factor_regions = [ff_region]
            else:
                raise TypeError(f"Invalid type for flux_factor_regions: {type(flux_factor_regions)}. "
                                "Must be list of 4 integers or None.")
            median2d = np.median(image3d, axis=0)
            flux_factor = compute_flux_factor(image3d, median2d, list_flux_factor_regions, _logger, interactive, debug)
            _logger.info("computed flux_factor set to %s", str(flux_factor))
        elif flux_factor.lower() == 'none':
            flux_factor = np.ones(num_images, dtype=float)
            if flux_factor_regions is not None:
                raise ValueError("Using flux_factor='none', but flux_factor_regions is provided. "
                                 "You must use flux_factor='auto' to use flux_factor_regions.")
        elif isinstance(ast.literal_eval(flux_factor), list):
            flux_factor = ast.literal_eval(flux_factor)
            if len(flux_factor) != num_images:
                raise ValueError(f"flux_factor must have the same length as the number of images ({num_images}).")
            if not all_valid_numbers(flux_factor):
                raise ValueError(f"All elements in flux_factor={flux_factor} must be valid numbers.")
            flux_factor = np.array(flux_factor, dtype=float)
        elif isinstance(ast.literal_eval(flux_factor), (float, int)):
            flux_factor = np.full(num_images, ast.literal_eval(flux_factor), dtype=float)
        else:
            raise ValueError(f"Invalid flux_factor string: {flux_factor}. Use 'auto' or 'none'.")
    elif isinstance(flux_factor, list):
        if len(flux_factor) != num_images:
            raise ValueError(f"flux_factor must have the same length as the number of images ({num_images}).")
        if not all_valid_numbers(flux_factor):
            raise ValueError(f"All elements in flux_factor={flux_factor} must be valid numbers.")
        flux_factor = np.array(flux_factor, dtype=float)
    else:
        raise ValueError(f"Invalid flux_factor value: {flux_factor}.")
    _logger.info("flux_factor: %s", str(flux_factor))
    if apply_flux_factor_to not in ['original', 'simulated']:
        raise ValueError(f"Invalid apply_flux_factor_to: {apply_flux_factor_to}. "
                         "Valid options are 'original' and 'simulated'.")
    _logger.info("apply_flux_factor_to: %s", apply_flux_factor_to)

    # Apply the flux factor to the input arrays if requested
    if apply_flux_factor_to == 'original':
        for i in range(num_images):
            image3d[i] /= flux_factor[i]

    # Define regions to be cleaned by computing a boolean mask
    # that is True for pixels not included in the regions to be skipped
    bool_to_be_cleaned = np.ones((naxis2, naxis1), dtype=bool)  # default is to clean all pixels
    if isinstance(regions_to_be_skipped, str):
        if regions_to_be_skipped.lower() == 'none':
            regions_to_be_skipped = None
        else:
            regions_to_be_skipped = ast.literal_eval(regions_to_be_skipped)
    if isinstance(regions_to_be_skipped, list):
        for region in regions_to_be_skipped:
            if isinstance(region, list):
                all_integers = all(isinstance(val, int) for val in region)
                if not all_integers:
                    raise TypeError(f"Invalid region_to_be_skipped: {region}. All elements must be integers.")
                if len(region) != 4:
                    raise ValueError(f"Invalid length for region_to_be_skipped: {region}. "
                                     "Must be a list of 4 integers [xmin, xmax, ymin, ymax].")
                dumreg = f"[{region[0]}:{region[1]}, {region[2]}:{region[3]}]"
                _logger.debug("defined region to be skipped: %s", dumreg)
                skip_region = tea.SliceRegion2D(dumreg, mode='fits', naxis1=naxis1, naxis2=naxis2)
                bool_to_be_cleaned[skip_region.python[0].start:skip_region.python[0].stop,
                                   skip_region.python[1].start:skip_region.python[1].stop] = False
            else:
                raise TypeError(f"Invalid type for region_to_be_skipped in the list: {type(region)}. "
                                "Must be a list of 4 integers.")

    # Compute minimum, maximum, median and mean along the first axis
    min2d = np.min(image3d, axis=0)
    max2d = np.max(image3d, axis=0)
    median2d = np.median(image3d, axis=0)
    mean2d = np.mean(image3d, axis=0)

    # Compute points for diagnostic diagram of mmcosmic method
    xplot = min2d.flatten()  # bias was already subtracted above
    yplot = median2d.flatten() - min2d.flatten()

    # Check that color_scale is valid
    if color_scale not in ['minmax', 'zscale']:
        raise ValueError(f"Invalid color_scale: {color_scale}. Valid options are 'minmax' and 'zscale'.")

    # Define the pixels to be flagged as CR
    if isinstance(pixels_to_be_flagged_as_cr, str):
        if pixels_to_be_flagged_as_cr.lower() == 'none':
            pixels_to_be_flagged_as_cr = None
        else:
            pixels_to_be_flagged_as_cr = ast.literal_eval(pixels_to_be_flagged_as_cr)
    if isinstance(pixels_to_be_flagged_as_cr, (list, tuple)):
        for p in pixels_to_be_flagged_as_cr:
            if (not isinstance(p, (list, tuple)) or len(p) != 2 or
                    not all(isinstance(item, int) for item in p)):
                raise ValueError(f"Invalid numbers in pixels_to_be_flagged_as_cr: {p}. "
                                 "Each pixel must be a tuple or list of two integers (X, Y).")
            if p[0] < 1 or p[0] > naxis1 or p[1] < 1 or p[1] > naxis2:
                raise ValueError(f"Pixel coordinates {p} in pixels_to_be_flagged_as_cr are out of bounds.")
            # ensure these pixels are cleaned, independently of being included
            # in regions_to_be_skipped
            bool_to_be_cleaned[p[1]-1, p[0]-1] = True
    elif pixels_to_be_flagged_as_cr is not None:
        raise TypeError(f"Invalid type for pixels_to_be_flagged_as_cr: {type(pixels_to_be_flagged_as_cr)}. "
                        "Must be a list of (x, y) tuples or None.")
    _logger.info("individual pixels to be initially flagged as CR: %s",
                 "None" if pixels_to_be_flagged_as_cr is None else str(pixels_to_be_flagged_as_cr))

    # Define the pixels to be ignored as CR
    # check that the provided pixels are valid
    if isinstance(pixels_to_be_ignored_as_cr, str):
        if pixels_to_be_ignored_as_cr.lower() == 'none':
            pixels_to_be_ignored_as_cr = None
        else:
            pixels_to_be_ignored_as_cr = ast.literal_eval(pixels_to_be_ignored_as_cr)
    if isinstance(pixels_to_be_ignored_as_cr, (list, tuple)):
        for p in pixels_to_be_ignored_as_cr:
            if (not isinstance(p, (list, tuple)) or len(p) != 2 or
                    not all(isinstance(item, int) for item in p)):
                raise ValueError(f"Invalid numbers in pixels_to_be_ignored_as_cr: {p}. "
                                 "Each pixel must be a tuple or list of two integers (X, Y).")
            if p[0] < 1 or p[0] > naxis1 or p[1] < 1 or p[1] > naxis2:
                raise ValueError(f"Pixel coordinates {p} in pixels_to_be_ignored_as_cr are out of bounds.")
            # ensure these pixels are not cleaned
            bool_to_be_cleaned[p[1]-1, p[0]-1] = False
    elif pixels_to_be_ignored_as_cr is not None:
        raise TypeError(f"Invalid type for pixels_to_be_ignored_as_cr: {type(pixels_to_be_ignored_as_cr)}. "
                        "Must be a list of (x, y) tuples or None.")
    _logger.info("individual pixels to be initially ignored as CR: %s",
                 "None" if pixels_to_be_ignored_as_cr is None else str(pixels_to_be_ignored_as_cr))

    # Define the pixels to be replaced by the median value when removing the CRs
    # check that the provided pixels are valid
    if isinstance(pixels_to_be_replaced_by_local_median, str):
        if pixels_to_be_replaced_by_local_median.lower() == 'none':
            pixels_to_be_replaced_by_local_median = None
    if isinstance(pixels_to_be_replaced_by_local_median, (list, tuple)):
        for p in pixels_to_be_replaced_by_local_median:
            if (not isinstance(p, (list, tuple)) or len(p) != 4 or
                    not all(isinstance(item, int) for item in p)):
                raise ValueError(f"Invalid numbers in pixels_to_be_replaced_by_local_median: {p}. "
                                 "Each pixel must be a tuple or list of four integers (X, Y, X_width, Y_width).")
            if p[0] < 1 or p[0] > naxis1 or p[1] < 1 or p[1] > naxis2:
                raise ValueError(f"Pixel coordinates {p} in pixels_to_be_replaced_by_local_median are out of bounds.")
            if p[2] % 2 == 0 or p[3] % 2 == 0 or p[2] < 1 or p[3] < 1:
                raise ValueError(f"Pixel {p}: X_width and Y_width in pixels_to_be_replaced_by_local_median "
                                 "must be odd integers >= 1.")
            if p[2] * p[3] < 3:
                raise ValueError(f"Pixel {p}: The area defined by X_width and Y_width in "
                                 "pixels_to_be_replaced_by_local_median must be >= 3.")
    elif pixels_to_be_replaced_by_local_median is not None:
        raise TypeError(f"Invalid type for pixels_to_be_replaced_by_local_median: "
                        f"{type(pixels_to_be_replaced_by_local_median)}. "
                        "Must be a list of (x, y, x_width, y_width) tuples or None.")
    _logger.info("pixels to be replaced by the median value when removing the CRs: %s",
                 "None" if pixels_to_be_replaced_by_local_median is None
                 else str(pixels_to_be_replaced_by_local_median))

    # These pixels to be replaced by the local median should not be
    # flagged as CR if they will be replaced by the local median anyway
    if pixels_to_be_replaced_by_local_median is not None:
        for p in pixels_to_be_replaced_by_local_median:
            if bool_to_be_cleaned[p[1]-1, p[0]-1]:
                _logger.warning("Pixel %s is set to be replaced by the local median "
                                "but it is also set to be cleaned as CR. "
                                "It will not be cleaned as CR but will be replaced by the local median.", str(p))
                bool_to_be_cleaned[p[1]-1, p[0]-1] = False
        _logger.info("updated pixels to be ignored as CR: %s",
                     "None" if pixels_to_be_ignored_as_cr is None else str(pixels_to_be_ignored_as_cr))

    # Log the input parameters
    if crmethod in ['mmcosmic', 'mm_lacosmic']:
        _logger.debug("mm_xy_offsets: %s", str(mm_xy_offsets) if mm_xy_offsets is not None else "None")
        _logger.debug("mm_crosscorr_region: %s", mm_crosscorr_region if mm_crosscorr_region is not None else "None")
        _logger.debug("mm_boundary_fit: %s", mm_boundary_fit if mm_boundary_fit is not None else "None")
        _logger.debug("mm_knots_splfit: %d", mm_knots_splfit)
        _logger.debug("mm_fixed points_in_boundary: %s",
                      str(mm_fixed_points_in_boundary) if mm_fixed_points_in_boundary is not None else "None")
        _logger.debug("mm_nsimulations: %d", mm_nsimulations)
        _logger.debug("mm_niter_boundary_extension: %d", mm_niter_boundary_extension)
        _logger.debug("mm_weight_boundary_extension: %f", mm_weight_boundary_extension)
        _logger.debug("mm_threshold: %s", mm_threshold if mm_threshold is not None else "None")
        _logger.debug("mm_minimum_max2d_rnoise: %f", mm_minimum_max2d_rnoise)
        _logger.debug("mm_seed: %s", str(mm_seed))

    if crmethod in ['lacosmic', 'mm_lacosmic']:
        # Check la_gain_apply
        if la_gain_apply is None:
            la_gain_apply = True
            _logger.warning("la_gain_apply for lacosmic not defined, assuming la_gain_apply=True")
        else:
            _logger.debug("la_gain_apply for lacosmic: %s", str(la_gain_apply))
        # Check la_sigclip
        if la_sigclip is None:
            _logger.warning("la_sigclip for lacosmic not defined, assuming la_sigclip=5.0")
            la_sigclip = 5.0
        else:
            _logger.debug("la_sigclip for lacosmic: %f", la_sigclip)
        # Check la_sigfrac
        if la_sigfrac is None:
            _logger.warning("la_sigfrac for lacosmic not defined, assuming la_sigfrac=0.3")
            la_sigfrac = 0.3
        else:
            _logger.debug("la_sigfrac for lacosmic: %f", la_sigfrac)
        # Check la_objlim
        if la_objlim is None:
            _logger.warning("la_objlim for lacosmic not defined, assuming la_objlim=5.0")
            la_objlim = 5.0
        else:
            _logger.debug("la_objlim for lacosmic: %f", la_objlim)
        # Check la_satlevel
        if la_satlevel is None:
            _logger.warning("la_satlevel for lacosmic not defined, assuming la_satlevel=None")
        else:
            _logger.debug("la_satlevel for lacosmic: %f", la_satlevel)
        # Check niter
        if la_niter is None:
            _logger.warning("la_niter for lacosmic not defined, assuming la_niter=4")
            la_niter = 4
        else:
            _logger.debug("la_niter for lacosmic: %d", la_niter)
        # Check la_sepmed
        if la_sepmed is None:
            _logger.warning("la_sepmed for lacosmic not defined, assuming la_sepmed=True")
            la_sepmed = True
        else:
            _logger.debug("la_sepmed for lacosmic: %s", str(la_sepmed))
        # Check la_cleantype
        if la_cleantype is None:
            raise ValueError("la_cleantype for lacosmic must be provided.")
        elif isinstance(la_cleantype, str):
            if la_cleantype not in VALID_LACOSMIC_CLEANTYPE:
                raise ValueError(f"la_cleantype must be one of {VALID_LACOSMIC_CLEANTYPE} or 'none'.")
        else:
            raise TypeError("la_cleantype must be a string.")
        # Check la_fsmode
        if la_fsmode not in ['median', 'convolve']:
            raise ValueError("la_fsmode must be 'median' or 'convolve'.")
        else:
            _logger.debug("la_fsmode for lacosmic: %s", la_fsmode)
        # Check la_psfmodel
        if la_psfmodel not in ['gauss', 'moffat', 'gaussx', 'gaussy', 'gaussxy']:
            raise ValueError("la_psfmodel must be 'gauss', 'moffat', 'gaussx', 'gaussy', or 'gaussxy'.")
        else:
            _logger.debug("la_psfmodel for lacosmic: %s", la_psfmodel)
        # Check la_psffwhm_x, la_psffwhm_y, la_psfsize
        if la_fsmode == 'convolve':
            if la_psffwhm_x is None or la_psffwhm_y is None or la_psfsize is None:
                raise ValueError("For la_fsmode='convolve', "
                                 "la_psffwhm_x, la_psffwhm_y, and la_psfsize must be provided.")
            else:
                _logger.debug("la_psffwhm_x for lacosmic: %f", la_psffwhm_x)
                _logger.debug("la_psffwhm_y for lacosmic: %f", la_psffwhm_y)
                _logger.debug("la_psfsize for lacosmic: %d", la_psfsize)
            if la_psfsize % 2 == 0 or la_psfsize < 3:
                raise ValueError("la_psfsize must be an odd integer >= 3.")
        # Check la_psfbeta
        if la_psfmodel == 'moffat':
            if la_psfbeta is None:
                raise ValueError("For la_psfmodel='moffat', la_psfbeta must be provided.")
            else:
                _logger.debug("la_psfbeta for lacosmic: %f", la_psfbeta)
        # Set la_verbose
        current_logging_level = logging.getLogger().getEffectiveLevel()
        if current_logging_level in [logging.WARNING, logging.ERROR, logging.CRITICAL]:
            la_verbose = False
        _logger.debug("la_verbose for lacosmic: %s", str(la_verbose))
        # Define dictionary with the parameters for cosmicray_lacosmic() function
        # Note: the "pssl" parameter is not used here because it was deprecated
        # in version 2.3.0 and will be removed in a future version.
        # The "pssl" keyword will be removed in ccdproc 3.0.
        # Use "inbkg" instead to have astroscrappy temporarily remove the background during processing.
        dict_la_params = {
            'gain': gain_scalar,
            'readnoise': rnoise_scalar,
            'sigclip': la_sigclip,
            'sigfrac': la_sigfrac,
            'objlim': la_objlim,
            'satlevel': la_satlevel * gain_scalar if la_satlevel is not None else None,  # in electrons!
            'niter': la_niter,
            'sepmed': la_sepmed,
            'cleantype': la_cleantype,
            'fsmode': la_fsmode,
            'psfmodel': la_psfmodel,
            'psffwhm': None,
            'psfsize': la_psfsize,
            'psfbeta': la_psfbeta,
            'verbose': la_verbose,
            'psfk': None,
            'inbkg': None,
            'invar': None
        }
        if la_psfmodel in ['gauss', 'moffat']:
            if la_psffwhm_x is None or la_psfsize is None:
                raise ValueError("For la_psfmodel='gauss' or 'moffat', "
                                 "la_psffwhm_x and la_psfsize must be provided.")
            dict_la_params['psffwhm'] = (la_psffwhm_x + la_psffwhm_y) / 2.0  # average FWHM
        elif la_psfmodel == 'gaussx':
            if la_psffwhm_x is None or la_psfsize is None:
                raise ValueError("For la_psfmodel='gaussx', "
                                 "la_psffwhm_x and la_psfsize must be provided.")
            dict_la_params['psffwhm'] = la_psffwhm_x
        elif la_psfmodel == 'gaussy':
            if la_psffwhm_y is None or la_psfsize is None:
                raise ValueError("For la_psfmodel='gaussy', "
                                 "la_psffwhm_y and la_psfsize must be provided.")
            dict_la_params['psffwhm'] = la_psffwhm_y
        elif la_psfmodel == 'gaussxy':
            dict_la_params['psffwhm'] = None  # not used in this case
            dict_la_params['psfk'] = gausskernel2d_elliptical(
                fwhm_x=la_psffwhm_x,
                fwhm_y=la_psffwhm_y,
                kernsize=la_psfsize
            )
        else:
            raise ValueError("la_psfmodel must be 'gauss', 'moffat', 'gaussx', 'gaussy', or 'gaussxy'.")

    _logger.info("dtype for output arrays: %s", dtype)
    _logger.info("dilation factor: %d", dilation)
    _logger.info("verify cosmic-ray detection: %s", verify_cr)
    _logger.info("semiwindow size for plotting coincident cosmic-ray pixels: %d", semiwindow)
    _logger.info("maximum number of coincident cosmic-ray pixels to plot: %d", maxplots)
    _logger.info("color scale for plots: %s", color_scale)

    if la_verbose:
        for key in dict_la_params.keys():
            if key == 'psfk':
                if dict_la_params[key] is None:
                    _logger.info("%s for lacosmic: None", key)
                else:
                    _logger.info("%s for lacosmic: array with shape %s", key, str(dict_la_params[key].shape))
            else:
                _logger.info("%s for lacosmic: %s", key, str(dict_la_params[key]))

    if rich_configured:
        _logger.info("[green]" + "-" * 79 + "[/green]")
        _logger.info("starting cosmic ray detection in [magenta]median2d[/magenta] image...")
    else:
        _logger.info("-" * 73)
        _logger.info("starting cosmic ray detection in median2d image...")

    if crmethod in ['lacosmic', 'mm_lacosmic']:
        # ---------------------------------------------------------------------
        # Detect residual cosmic rays in the median2d image using the
        # Laplacian edge detection method from ccdproc. This only works if gain and
        # rnoise are constant values (scalars).
        # ---------------------------------------------------------------------
        _logger.info(f"detecting cosmic rays in median2d image using {rlabel_lacosmic}...")
        if gain_scalar is None or rnoise_scalar is None:
            raise ValueError("gain and rnoise must be constant values (scalars) when using crmethod='lacosmic'.")
        median2d_lacosmic, flag_la = decorated_cosmicray_lacosmic(
            ccd=median2d,
            **{key: value for key, value in dict_la_params.items() if value is not None}
        )
        _logger.info("pixels flagged as cosmic rays by %s: %d (%08.4f%%)",
                     rlabel_lacosmic, np.sum(flag_la), np.sum(flag_la) / flag_la.size * 100)
        flag_la = np.logical_and(flag_la, bool_to_be_cleaned)
        flag_la = flag_la.flatten()
        if crmethod == 'lacosmic':
            xplot_boundary = None
            yplot_boundary = None
            mm_threshold = None
            flag_sb = np.zeros_like(flag_la, dtype=bool)
    else:
        median2d_lacosmic = None

    if crmethod in ['mmcosmic', 'mm_lacosmic']:
        # ---------------------------------------------------------------------
        # Detect cosmic rays in the median2d image using the numerically
        # derived boundary.
        # ---------------------------------------------------------------------
        # Define mm_fixed_points_in_boundary
        _logger.info("detecting cosmic rays in median2d image using %s...", rlabel_mmcosmic)
        if isinstance(mm_fixed_points_in_boundary, str):
            if mm_fixed_points_in_boundary.lower() == 'none':
                mm_fixed_points_in_boundary = None
        if mm_fixed_points_in_boundary is None:
            if mm_boundary_fit == 'piecewise':
                raise ValueError("For mm_boundary_fit='piecewise', "
                                 "mm_fixed_points_in_boundary must be provided.")
        else:
            mm_fixed_points_in_boundary = list(eval(str(mm_fixed_points_in_boundary)))
            x_mm_fixed_points_in_boundary = []
            y_mm_fixed_points_in_boundary = []
            w_mm_fixed_points_in_boundary = []
            for item in mm_fixed_points_in_boundary:
                if not (isinstance(item, (list, tuple)) and len(item) in [2, 3]):
                    raise ValueError("Each item in mm_fixed_points_in_boundary must be a list or tuple of "
                                     "2 or 3 elements: (x, y) or (x, y, weight).")
                if not all_valid_numbers(item):
                    raise ValueError(f"All elements in mm_fixed_points_in_boundary={mm_fixed_points_in_boundary} "
                                     "must be valid numbers.")
                if len(item) == 2:
                    x_mm_fixed_points_in_boundary.append(float(item[0]))
                    y_mm_fixed_points_in_boundary.append(float(item[1]))
                    w_mm_fixed_points_in_boundary.append(10000)
                else:
                    x_mm_fixed_points_in_boundary.append(float(item[0]))
                    y_mm_fixed_points_in_boundary.append(float(item[1]))
                    w_mm_fixed_points_in_boundary.append(float(item[2]))
            x_mm_fixed_points_in_boundary = np.array(x_mm_fixed_points_in_boundary, dtype=float)
            y_mm_fixed_points_in_boundary = np.array(y_mm_fixed_points_in_boundary, dtype=float)
            w_mm_fixed_points_in_boundary = np.array(w_mm_fixed_points_in_boundary, dtype=float)

        if mm_boundary_fit is None:
            raise ValueError(f"mm_boundary_fit is None and must be one of {VALID_BOUNDARY_FITS}.")
        elif mm_boundary_fit not in VALID_BOUNDARY_FITS:
            raise ValueError(f"Invalid mm_boundary_fit: {mm_boundary_fit}. Valid options are {VALID_BOUNDARY_FITS}.")
        if mm_boundary_fit == 'piecewise':
            if mm_fixed_points_in_boundary is None:
                raise ValueError("For mm_boundary_fit='piecewise', "
                                 "mm_fixed_points_in_boundary must be provided.")
            elif len(x_mm_fixed_points_in_boundary) < 2:
                raise ValueError("For mm_boundary_fit='piecewise', "
                                 "at least two fixed points must be provided in mm_fixed_points_in_boundary.")

        # Compute offsets between each single exposure and the median image
        if isinstance(mm_xy_offsets, str):
            if mm_xy_offsets.lower() == 'none':
                mm_xy_offsets = None
        if mm_xy_offsets is not None and mm_crosscorr_region is not None:
            raise ValueError("You can only provide one of mm_xy_offsets or mm_crosscorr_region, not both.")
        shift_images = False
        if mm_xy_offsets is not None:
            if isinstance(mm_xy_offsets, str):
                mm_xy_offsets = ast.literal_eval(mm_xy_offsets)
            if isinstance(mm_xy_offsets, list):
                if len(mm_xy_offsets) != num_images:
                    raise ValueError(f"mm_xy_offsets must have the same length as the number of images ({num_images}).")
                for offset in mm_xy_offsets:
                    if (not isinstance(offset, (list, tuple)) or len(offset) != 2 or
                            not all_valid_numbers(offset)):
                        raise ValueError(f"Invalid offset in mm_xy_offsets: {offset}. "
                                         "Each offset must be a tuple or list of two numbers (x_offset, y_offset).")
                list_yx_offsets = [(float(offset[1]), float(offset[0])) for offset in mm_xy_offsets]
                for i, yx_offsets in enumerate(list_yx_offsets):
                    _logger.info("provided offsets for image %d: y=%+f, x=%+f", i+1, yx_offsets[0], yx_offsets[1])
            else:
                raise TypeError(f"Invalid type for mm_xy_offsets: {type(mm_xy_offsets)}. "
                                "Must be list of [x_offset, y_offset)].")
            shift_images = True
        else:
            if isinstance(mm_crosscorr_region, str):
                if mm_crosscorr_region.lower() == 'none':
                    mm_crosscorr_region = None
                else:
                    mm_crosscorr_region = ast.literal_eval(mm_crosscorr_region)
            if isinstance(mm_crosscorr_region, list):
                all_integers = all(isinstance(val, int) for val in mm_crosscorr_region)
                if not all_integers:
                    raise TypeError(f"Invalid mm_crosscorr_region: {mm_crosscorr_region}. "
                                    "All elements must be integers.")
                if len(mm_crosscorr_region) != 4:
                    raise ValueError(f"Invalid length for mm_crosscorr_region: {mm_crosscorr_region}. "
                                     "Must be a list of 4 integers [xmin, xmax, ymin, ymax].")
                dumreg = f"[{mm_crosscorr_region[0]}:{mm_crosscorr_region[1]}, " + \
                         f"{mm_crosscorr_region[2]}:{mm_crosscorr_region[3]}]"
                _logger.debug("defined mm_crosscorr_region: %s", dumreg)
                crossregion = tea.SliceRegion2D(dumreg, mode='fits', naxis1=naxis1, naxis2=naxis2)
                if crossregion.area() < 100:
                    raise ValueError("The area of mm_crosscorr_region must be at least 100 pixels.")
                shift_images = True
            elif mm_crosscorr_region is None:
                crossregion = None
            else:
                raise TypeError(f"Invalid type for mm_crosscorr_region: {type(mm_crosscorr_region)}. "
                                "Must be list of 4 integers or None.")
            list_yx_offsets = []
            for i in range(num_images):
                if crossregion is None:
                    list_yx_offsets.append((0.0, 0.0))
                else:
                    reference_image = median2d[crossregion.python]
                    moving_image = image3d[i][crossregion.python]
                    _logger.info("computing offsets for image %d using cross-correlation...", i+1)
                    yx_offsets, _, _ = phase_cross_correlation(
                        reference_image=reference_image,
                        moving_image=moving_image,
                        upsample_factor=100,
                        normalization=None  # use None to avoid artifacts with images with many cosmic rays
                    )
                    yx_offsets[0] *= -1  # invert sign
                    yx_offsets[1] *= -1  # invert sign
                    _logger.info("offsets for image %d: y=%+f, x=%+f", i+1, yx_offsets[0], yx_offsets[1])

                    def on_key_cross(event):
                        if event.key == 'x':
                            _logger.info("Exiting program as per user request ('x' key pressed).")
                            plt.close(fig)
                            sys.exit(0)

                    fig, axarr = plt.subplots(nrows=2, ncols=2, sharex=True, sharey=True, figsize=(6.4*1.5, 4.8*1.5))
                    fig.canvas.mpl_connect('key_press_event', lambda event: on_key_cross(event))
                    axarr = axarr.flatten()
                    vmin = np.min(reference_image)
                    vmax = np.max(reference_image)
                    extent = [crossregion.fits[0].start - 0.5, crossregion.fits[0].stop + 0.5,
                              crossregion.fits[1].start - 0.5, crossregion.fits[1].stop + 0.5]
                    tea.imshow(fig, axarr[0], reference_image, vmin=vmin, vmax=vmax,
                               extent=extent, aspect='auto', title='Median')
                    tea.imshow(fig, axarr[1], moving_image, vmin=vmin, vmax=vmax,
                               extent=extent, aspect='auto', title=f'Image {i+1}')
                    shifted_image2d = shift_image2d(
                        moving_image,
                        xoffset=-yx_offsets[1],
                        yoffset=-yx_offsets[0],
                        resampling=2
                    )
                    dumdiff1 = reference_image - moving_image
                    dumdiff2 = reference_image - shifted_image2d
                    vmin = np.percentile(dumdiff1, 5)
                    vmax = np.percentile(dumdiff2, 95)
                    tea.imshow(fig, axarr[2], dumdiff1, vmin=vmin, vmax=vmax,
                               extent=extent, aspect='auto', title=f'Median - Image {i+1}')
                    tea.imshow(fig, axarr[3], dumdiff2, vmin=vmin, vmax=vmax,
                               extent=extent, aspect='auto', title=f'Median - Shifted Image {i+1}')
                    plt.tight_layout()
                    png_filename = f'xyoffset_crosscorr_{i+1}.png'
                    _logger.info(f"saving {png_filename}")
                    plt.savefig(png_filename, dpi=150)
                    if interactive:
                        _logger.info("Entering interactive mode (press 'q' to close figure, 'x' to quit program)")
                        plt.show()
                    plt.close(fig)
                    list_yx_offsets.append(yx_offsets)

        # Estimate limits for the diagnostic plot
        rng = np.random.default_rng(mm_seed)  # Random number generator for reproducibility
        xdiag_min, xdiag_max, ydiag_min, ydiag_max = estimate_diagnostic_limits(
            rng=rng,
            gain=np.median(gain),  # Use median value to simplify the computation
            rnoise=np.median(rnoise),  # Use median value to simplify the computation
            maxvalue=np.max(min2d),
            num_images=num_images,
            npixels=10000
        )
        if np.min(xplot) < xdiag_min:
            xdiag_min = np.min(xplot)
        if np.max(xplot) > xdiag_max:
            xdiag_max = np.max(xplot)
        if mm_fixed_points_in_boundary is not None:
            if np.max(y_mm_fixed_points_in_boundary) > ydiag_max:
                ydiag_max = np.max(y_mm_fixed_points_in_boundary)
        if shift_images:
            ydiag_max *= 4.0  # Add 300% margin to the maximum y limit
        else:
            ydiag_max *= 2.0  # Add 100% margin to the maximum y limit
        _logger.debug("xdiag_min=%f", xdiag_min)
        _logger.debug("ydiag_min=%f", ydiag_min)
        _logger.debug("xdiag_max=%f", xdiag_max)
        _logger.debug("ydiag_max=%f", ydiag_max)

        # Define binning for the diagnostic plot
        nbins_xdiag = 100
        nbins_ydiag = 100
        bins_xdiag = np.linspace(xdiag_min, xdiag_max, nbins_xdiag + 1)
        bins_ydiag = np.linspace(0, ydiag_max, nbins_ydiag + 1)
        xcbins = (bins_xdiag[:-1] + bins_xdiag[1:]) / 2
        ycbins = (bins_ydiag[:-1] + bins_ydiag[1:]) / 2

        # Create a 2D histogram for the diagnostic plot, using
        # integers to avoid rounding errors
        hist2d_accummulated = np.zeros((nbins_ydiag, nbins_xdiag), dtype=int)
        lam = median2d.copy()
        lam[lam < 0] = 0  # Avoid negative values
        lam3d = np.zeros((num_images, naxis2, naxis1))
        if apply_flux_factor_to == 'original':
            flux_factor_for_simulated = np.ones(num_images, dtype=float)
        elif apply_flux_factor_to == 'simulated':
            flux_factor_for_simulated = flux_factor
        else:
            raise ValueError(f"Invalid apply_flux_factor_to: {apply_flux_factor_to}. "
                             "Valid options are 'original' and 'simulated'.")
        _logger.info("flux factor for simulated images: %s", str(flux_factor_for_simulated))
        if not shift_images:
            _logger.info("assuming no offsets between images")
            for i in range(num_images):
                lam3d[i] = lam / flux_factor_for_simulated[i]
        else:
            _logger.info("xy-shifting median2d to speed up simulations...")
            for i in range(num_images):
                _logger.info("shifted image %d/%d -> delta_y=%+f, delta_x=%+f",
                             i + 1, num_images, list_yx_offsets[i][0], list_yx_offsets[i][1])
                # apply offsets to the median image to simulate the expected individual exposures
                lam3d[i] = shift_image2d(lam,
                                         xoffset=list_yx_offsets[i][1],
                                         yoffset=list_yx_offsets[i][0],
                                         resampling=2) / flux_factor_for_simulated[i]
        _logger.info("computing simulated 2D histogram...")
        for k in range(mm_nsimulations):
            time_ini = datetime.now()
            image3d_simul = np.zeros((num_images, naxis2, naxis1))
            for i in range(num_images):
                # convert from ADU to electrons to apply Poisson noise, and then back to ADU
                image3d_simul[i] = rng.poisson(lam=lam3d[i] * gain).astype(float) / gain
                # add readout noise in ADU
                image3d_simul[i] += rng.normal(loc=0, scale=rnoise)
            min2d_simul = np.min(image3d_simul, axis=0)
            median2d_simul = np.median(image3d_simul, axis=0)
            xplot_simul = min2d_simul.flatten()
            yplot_simul = median2d_simul.flatten() - min2d_simul.flatten()
            hist2d, edges = np.histogramdd(
                sample=(yplot_simul, xplot_simul),
                bins=(bins_ydiag, bins_xdiag)
            )
            hist2d_accummulated += hist2d.astype(int)
            time_end = datetime.now()
            _logger.info("simulation %d/%d, time elapsed: %s", k + 1, mm_nsimulations, time_end - time_ini)
        # Average the histogram over the number of simulations
        hist2d_accummulated = hist2d_accummulated.astype(float) / mm_nsimulations
        vmin = np.min(hist2d_accummulated[hist2d_accummulated > 0])
        if vmin == 0:
            vmin = 1
        vmax = np.max(hist2d_accummulated)
        cmap1 = plt.get_cmap('cividis_r')
        cmap2 = plt.get_cmap('viridis')
        n_colors = 256
        n_colors2 = int((np.log10(vmax) - np.log10(1.0)) / (np.log10(vmax) - np.log10(vmin)) * n_colors)
        n_colors2 += 1
        if n_colors2 > n_colors:
            n_colors2 = n_colors
        if n_colors2 < n_colors:
            n_colors1 = n_colors - n_colors2
        else:
            n_colors1 = 0
        colors1 = cmap1(np.linspace(0, 1, n_colors1))
        colors2 = cmap2(np.linspace(0, 1, n_colors2))
        combined_colors = np.vstack((colors1, colors2))
        combined_cmap = LinearSegmentedColormap.from_list('combined_cmap', combined_colors)
        norm = LogNorm(vmin=vmin, vmax=vmax)

        def on_key_2dhist(event):
            if event.key == 'x':
                _logger.info("Exiting program as per user request ('x' key pressed).")
                plt.close(fig)
                sys.exit(0)

        fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(12.1, 5.5))
        fig.canvas.mpl_connect('key_press_event', lambda event: on_key_2dhist(event))
        # Display 2D histogram of the simulated data
        extent = [bins_xdiag[0], bins_xdiag[-1], bins_ydiag[0], bins_ydiag[-1]]
        tea.imshow(fig, ax1, hist2d_accummulated, norm=norm, extent=extent,
                   aspect='auto', cblabel='Number of pixels', cmap=combined_cmap)
        # Display 2D histogram of the original data
        hist2d_original, edges = np.histogramdd(
            sample=(yplot, xplot),
            bins=(bins_ydiag, bins_xdiag)
        )
        tea.imshow(fig, ax2, hist2d_original, norm=norm, extent=extent,
                   aspect='auto', cblabel='Number of pixels', cmap=combined_cmap)

        # Determine the detection boundary for coincident cosmic-ray detection
        _logger.info("computing numerical boundary for coincident cosmic-ray detection...")
        xboundary = []
        yboundary = []
        for i in range(nbins_xdiag):
            fsum = np.sum(hist2d_accummulated[:, i])
            if fsum > 0:
                pdensity = hist2d_accummulated[:, i] / fsum
                perc = (1 - (1 / mm_nsimulations) / fsum)
                p = np.interp(perc, np.cumsum(pdensity), np.arange(nbins_ydiag))
                xboundary.append(xcbins[i])
                yboundary.append(ycbins[int(p + 0.5)])
        xboundary = np.array(xboundary)
        yboundary = np.array(yboundary)
        ax1.plot(xboundary, yboundary, 'r+')
        boundaryfit = None  # avoid flake8 warning
        if mm_boundary_fit == 'spline':
            for iterboundary in range(mm_niter_boundary_extension + 1):
                wboundary = np.ones_like(xboundary, dtype=float)
                if iterboundary == 0:
                    label = 'initial spline fit'
                else:
                    wboundary[yboundary > boundaryfit(xboundary)] = mm_weight_boundary_extension**iterboundary
                    label = f'Iteration {iterboundary}'
                if mm_fixed_points_in_boundary is None:
                    xboundary_fit = xboundary
                    yboundary_fit = yboundary
                    wboundary_fit = wboundary
                else:
                    wboundary_max = np.max(wboundary)
                    xboundary_fit = np.concatenate((xboundary, x_mm_fixed_points_in_boundary))
                    yboundary_fit = np.concatenate((yboundary, y_mm_fixed_points_in_boundary))
                    wboundary_fit = np.concatenate((wboundary, w_mm_fixed_points_in_boundary * wboundary_max))
                isort = np.argsort(xboundary_fit)
                boundaryfit, knots = spline_positive_derivative(
                    x=xboundary_fit[isort],
                    y=yboundary_fit[isort],
                    w=wboundary_fit[isort],
                    n_total_knots=mm_knots_splfit,
                )
                ydum = boundaryfit(xcbins)
                ydum[xcbins < knots[0]] = boundaryfit(knots[0])
                ydum[xcbins > knots[-1]] = boundaryfit(knots[-1])
                ax1.plot(xcbins, ydum, '-', color=f'C{iterboundary}', label=label)
                ax1.plot(knots, boundaryfit(knots), 'o', color=f'C{iterboundary}', markersize=4)
        elif mm_boundary_fit == 'piecewise':
            boundaryfit = define_piecewise_linear_function(
                xarray=x_mm_fixed_points_in_boundary,
                yarray=y_mm_fixed_points_in_boundary
            )
            ax1.plot(xcbins, boundaryfit(xcbins), 'r-', label='Piecewise linear fit')
        else:
            raise ValueError(f"Invalid mm_boundary_fit: {mm_boundary_fit}. Valid options are {VALID_BOUNDARY_FITS}.")

        if mm_threshold is None:
            # Use the minimum value of the boundary as the mm_threshold
            mm_threshold = np.min(yplot_boundary)
            _logger.info("updated mm_threshold for cosmic-ray detection: %f", mm_threshold)

        # Apply the criterium to detect coincident cosmic-ray pixels
        flag1 = yplot > boundaryfit(xplot)
        flag2 = yplot > mm_threshold
        flag_sb = np.logical_and(flag1, flag2)
        flag3 = max2d.flatten() > mm_minimum_max2d_rnoise * rnoise.flatten()
        flag_sb = np.logical_and(flag_sb, flag3)
        flag_sb = np.logical_and(flag_sb, bool_to_be_cleaned.flatten())
        _logger.info("pixels flagged as cosmic rays by %s: %d (%08.4f%%)",
                     rlabel_mmcosmic, np.sum(flag_sb), np.sum(flag_sb) / flag_sb.size * 100)
        if crmethod == 'mmcosmic':
            flag_la = np.zeros_like(flag_sb, dtype=bool)

        # Plot the results
        if mm_fixed_points_in_boundary is not None:
            ax1.plot(x_mm_fixed_points_in_boundary, y_mm_fixed_points_in_boundary, 'ms', markersize=6, alpha=0.5,
                     label='Fixed points')
        ax1.set_xlabel(r'min2d $-$ bias')
        ax1.set_ylabel(r'median2d $-$ min2d')
        ax1.set_title(f'Simulated data (mm_nsimulations = {mm_nsimulations})')
        if mm_niter_boundary_extension > 1:
            ax1.legend(loc=4)
        xplot_boundary = np.linspace(xdiag_min, xdiag_max, 100)
        yplot_boundary = boundaryfit(xplot_boundary)
        if mm_boundary_fit == 'spline':
            # For spline fit, force the boundary to be constant outside the knots
            yplot_boundary[xplot_boundary < knots[0]] = boundaryfit(knots[0])
            yplot_boundary[xplot_boundary > knots[-1]] = boundaryfit(knots[-1])
        ax2.plot(xplot_boundary, yplot_boundary, 'r-', label='Detection boundary')
        if mm_fixed_points_in_boundary is not None:
            ax2.plot(x_mm_fixed_points_in_boundary, y_mm_fixed_points_in_boundary, 'ms', markersize=6, alpha=0.5,
                     label='Fixed points')
        ax2.set_xlim(xdiag_min, xdiag_max)
        ax2.set_ylim(ax1.get_ylim())
        ax2.set_xlabel(ax1.get_xlabel())
        ax2.set_ylabel(ax1.get_ylabel())
        ax2.set_title('Original data')
        ax2.legend(loc=4)
        plt.tight_layout()
        png_filename = 'diagnostic_histogram2d.png'
        _logger.info(f"saving {png_filename}")
        plt.savefig(png_filename, dpi=150)
        if interactive:
            _logger.info("Entering interactive mode (press 'q' to close figure, 'x' to quit program)")
            plt.show()
        plt.close(fig)

    # Define the final cosmic ray flag
    if flag_la is None and flag_sb is None:
        raise RuntimeError("Both flag_la and flag_sb are None. This should never happen.")
    elif flag_la is None:
        flag = flag_sb
        flag_integer = 2 * flag_sb.astype(np.uint8)
    elif flag_sb is None:
        flag = flag_la
        flag_integer = 3 * flag_la.astype(np.uint8)
    else:
        # Combine the flags from lacosmic and mmcosmic
        flag = np.logical_or(flag_la, flag_sb)
        flag_integer = 2 * flag_sb.astype(np.uint8) + 3 * flag_la.astype(np.uint8)
        sdum = str(np.sum(flag))
        cdum = f"{np.sum(flag):{len(sdum)}d}"
        _logger.info("pixels flagged as cosmic rays by "
                     "%s or  %s: %s (%08.4f%%)", rlabel_lacosmic, rlabel_mmcosmic, cdum,
                     np.sum(flag) / flag.size * 100)
        cdum = f"{np.sum(flag_integer == 3):{len(sdum)}d}"
        _logger.info("pixels flagged as cosmic rays by "
                     "%s only........: %s (%08.4f%%)", rlabel_lacosmic, cdum,
                     np.sum(flag_integer == 3) / flag.size * 100)
        cdum = f"{np.sum((flag_integer == 2)):{len(sdum)}d}"
        _logger.info("pixels flagged as cosmic rays by "
                     "%s only........: %s (%08.4f%%)", rlabel_mmcosmic, cdum,
                     np.sum(flag_integer == 2) / flag.size * 100)
        cdum = f"{np.sum((flag_integer == 5)):{len(sdum)}d}"
        _logger.info("pixels flagged as cosmic rays by "
                     "%s and %s: %s (%08.4f%%)", rlabel_lacosmic, rlabel_mmcosmic, cdum,
                     np.sum((flag_integer == 5)) / flag.size * 100)
    flag = flag.reshape((naxis2, naxis1))
    flag_integer = flag_integer.reshape((naxis2, naxis1))
    flag_integer[flag_integer == 5] = 4  # pixels flagged by both methods are set to 4

    # Show diagnostic plot for the cosmic ray detection
    _logger.info("generating diagnostic plot for MEDIANCR...")
    ylabel = r'median2d $-$ min2d'
    diagnostic_plot(xplot, yplot, xplot_boundary, yplot_boundary, flag_la, flag_sb,
                    mm_threshold, ylabel, interactive,
                    target2d=median2d, target2d_name='median2d',
                    min2d=min2d, mean2d=mean2d, image3d=image3d,
                    _logger=_logger, png_filename='diagnostic_mediancr.png')

    # Check if any cosmic ray was detected
    if not np.any(flag):
        _logger.info("no coincident cosmic-ray pixels detected.")
        mask_mediancr = np.zeros_like(median2d, dtype=bool)
    else:
        _logger.info("coincident cosmic-ray pixels detected...")
        # Use the integer version of the flag for dilation
        if dilation > 0:
            structure = ndimage.generate_binary_structure(2, 2)
            flag_integer_dilated = ndimage.binary_dilation(
                flag_integer,
                structure=structure,
                iterations=dilation
            ).astype(np.uint8)
            sdum = str(np.sum(flag_integer_dilated > 0))
            cdum = f"{np.sum(flag_integer > 0):{len(sdum)}d}"
            _logger.info("before dilation: %s pixels flagged as coincident cosmic-ray pixels", cdum)
            cdum = f"{np.sum(flag_integer_dilated > 0):{len(sdum)}d}"
            _logger.info("after dilation : %s pixels flagged as coincident cosmic-ray pixels", cdum)
        else:
            flag_integer_dilated = flag_integer
            _logger.info("no dilation applied: %d pixels flagged as coincident cosmic-ray pixels",
                         np.sum(flag_integer > 0))
        # Set the pixels that were originally flagged as cosmic rays
        # to the integer value before dilation (this is to distinguish them
        # from the pixels that were dilated,which will be set to 1)
        flag_integer_dilated[flag] = flag_integer[flag]
        # Compute mask
        mask_mediancr = flag_integer_dilated > 0
        # Fix the median2d array by replacing the flagged pixels with the minimum value
        # of the corresponding pixel in the input arrays
        median2d_corrected = median2d.copy()
        if use_lamedian:
            median2d_corrected[mask_mediancr] = median2d_lacosmic[mask_mediancr]
        else:
            median2d_corrected[mask_mediancr] = min2d[mask_mediancr]
        # Label the connected pixels as individual cosmic rays
        labels_cr, number_cr = ndimage.label(flag_integer_dilated > 0)
        _logger.info("number of grouped cosmic-ray pixels detected: %d", number_cr)
        display_detected_cr(
            num_images=num_images,
            image3d=image3d,
            median2d=median2d,
            median2d_corrected=median2d_corrected,
            flag_integer_dilated=flag_integer_dilated,
            labels_cr=labels_cr,
            number_cr=number_cr,
            mask_mediancr=mask_mediancr,
            list_mask_single_exposures=None,
            mask_all_singles=None,
            semiwindow=semiwindow,
            maxplots=maxplots,
            verify_cr=verify_cr,
            color_scale=color_scale,
            _logger=_logger
        )

    # Generate list of HDUs with masks
    hdu_mediancr = fits.ImageHDU(mask_mediancr.astype(np.uint8), name='MEDIANCR')
    list_hdu_masks = [hdu_mediancr]

    # Apply the same algorithm but now with mean2d and with each individual array
    for i, target2d in enumerate([mean2d] + list_arrays):
        if i == 0:
            target2d_name = 'mean2d'
        else:
            target2d_name = f'single exposure #{i}'
        if rich_configured:
            _logger.info("[green]" + "-" * 79 + "[/green]")
            _logger.info(f"starting cosmic ray detection in [magenta]{target2d_name}[/magenta] image...")
        else:
            _logger.info("-" * 73)
            _logger.info(f"starting cosmic ray detection in {target2d_name} image...")
        if crmethod in ['lacosmic', 'mm_lacosmic']:
            _logger.info(f"detecting cosmic rays in {target2d_name} using {rlabel_lacosmic}...")
            array_lacosmic, flag_la = decorated_cosmicray_lacosmic(
                ccd=target2d,
                **{key: value for key, value in dict_la_params.items() if value is not None}
            )
            flag_la = np.logical_and(flag_la, bool_to_be_cleaned)
            flag_la = flag_la.flatten()
            if crmethod == 'lacosmic':
                xplot_boundary = None
                yplot_boundary = None
                mm_threshold = None
                flag_sb = np.zeros_like(flag_la, dtype=bool)
        if crmethod in ['mmcosmic', 'mm_lacosmic']:
            _logger.info(f"detecting cosmic rays in {target2d_name} using {rlabel_mmcosmic}...")
            xplot = min2d.flatten()
            yplot = target2d.flatten() - min2d.flatten()
            flag1 = yplot > boundaryfit(xplot)
            flag2 = yplot > mm_threshold
            flag_sb = np.logical_and(flag1, flag2)
            flag3 = max2d.flatten() > mm_minimum_max2d_rnoise * rnoise.flatten()
            flag_sb = np.logical_and(flag_sb, flag3)
            flag_sb = np.logical_and(flag_sb, bool_to_be_cleaned.flatten())
            if crmethod == 'mmcosmic':
                flag_la = np.zeros_like(flag_sb, dtype=bool)
        # For the mean2d mask, force the flag to be True if the pixel
        # was flagged as a coincident cosmic-ray pixel when using the median2d array
        # (this is to ensure that all pixels flagged in MEDIANCR are also
        # flagged in MEANCRT)
        if i == 0:
            flag_la = np.logical_or(flag_la, list_hdu_masks[0].data.astype(bool).flatten())
            flag_sb = np.logical_or(flag_sb, list_hdu_masks[0].data.astype(bool).flatten())
        # For the individual array masks, force the flag to be True if the pixel
        # is flagged both in the individual exposure and in the mean2d array
        if i > 0:
            flag_la = np.logical_and(flag_la, list_hdu_masks[1].data.astype(bool).flatten())
            flag_sb = np.logical_and(flag_sb, list_hdu_masks[1].data.astype(bool).flatten())
        sflag_la = str(np.sum(flag_la))
        sflag_sb = str(np.sum(flag_sb))
        smax = max(len(sflag_la), len(sflag_sb))
        _logger.info("pixels flagged as cosmic rays by "
                     "%s: %s (%08.4f%%)", rlabel_lacosmic, f"{np.sum(flag_la):{smax}d}",
                     np.sum(flag_la) / flag_la.size * 100)
        _logger.info("pixels flagged as cosmic rays by "
                     "%s: %s (%08.4f%%)", rlabel_mmcosmic, f"{np.sum(flag_sb):{smax}d}",
                     np.sum(flag_sb) / flag_sb.size * 100)
        if i == 0:
            _logger.info("generating diagnostic plot for MEANCRT...")
            png_filename = 'diagnostic_meancr.png'
            ylabel = r'mean2d $-$ min2d'
        else:
            _logger.info(f"generating diagnostic plot for CRMASK{i}...")
            png_filename = f'diagnostic_crmask{i}.png'
            ylabel = f'array{i}' + r' $-$ min2d'
        interactive_eff = interactive and debug
        diagnostic_plot(xplot, yplot, xplot_boundary, yplot_boundary, flag_la, flag_sb,
                        mm_threshold, ylabel, interactive_eff,
                        target2d=target2d, target2d_name=target2d_name,
                        min2d=min2d, mean2d=mean2d, image3d=image3d,
                        _logger=_logger, png_filename=png_filename)
        flag = np.logical_or(flag_la, flag_sb)
        flag = flag.reshape((naxis2, naxis1))
        flag_integer = flag.astype(np.uint8)
        if dilation > 0:
            structure = ndimage.generate_binary_structure(2, 2)
            flag_integer_dilated = ndimage.binary_dilation(
                flag_integer,
                structure=structure,
                iterations=dilation
            ).astype(np.uint8)
            sdum = str(np.sum(flag_integer_dilated))
            cdum = f"{np.sum(flag_integer):{len(sdum)}d}"
            _logger.info("before dilation: %s pixels flagged as cosmic rays", cdum)
            cdum = f"{np.sum(flag_integer_dilated):{len(sdum)}d}"
            _logger.info("after dilation : %s pixels flagged as cosmic rays", cdum)
        else:
            flag_integer_dilated = flag_integer
            _logger.info("no dilation applied: %d pixels flagged as cosmic rays", np.sum(flag_integer))
        flag_integer_dilated[flag] = 2
        # Compute mask
        mask = flag_integer_dilated > 0
        if i == 0:
            name = 'MEANCRT'
        else:
            name = f'CRMASK{i}'
        hdu_mask = fits.ImageHDU(mask.astype(np.uint8), name=name)
        list_hdu_masks.append(hdu_mask)

    if median2d_lacosmic is not None:
        hdu_median2d_lacosmic = fits.ImageHDU(median2d_lacosmic.astype(np.float32), name='LAMEDIAN')
        list_hdu_masks.append(hdu_median2d_lacosmic)

    # Find problematic cosmic-ray pixels (those masked in all individual CRMASKi)
    mask_all_singles = np.ones((naxis2, naxis1), dtype=bool)
    for hdu in list_hdu_masks[2:]:
        mask_all_singles = np.logical_and(mask_all_singles, hdu.data.astype(bool))
    problematic_pixels = np.argwhere(mask_all_singles)
    if rich_configured:
        _logger.info("[green]" + "-" * 79 + "[/green]")
    else:
        _logger.info("-" * 73)
    _logger.info("number of problematic cosmic-ray pixels masked in all individual CRMASKi: %d",
                 len(problematic_pixels))
    if len(problematic_pixels) > 0:
        # Label the connected problematic pixels as individual problematic cosmic rays
        labels_cr, number_cr = ndimage.label(mask_all_singles)
        _logger.info("number of grouped problematic cosmic-ray pixels: %d", number_cr)
        display_detected_cr(
            num_images=num_images,
            image3d=image3d,
            median2d=median2d,
            median2d_corrected=median2d_corrected,
            flag_integer_dilated=flag_integer_dilated,
            labels_cr=labels_cr,
            number_cr=number_cr,
            mask_mediancr=mask_mediancr,
            list_mask_single_exposures=[hdu.data for hdu in list_hdu_masks[2:2+num_images]],
            mask_all_singles=mask_all_singles,
            semiwindow=semiwindow,
            maxplots=maxplots,
            verify_cr=False,
            color_scale=color_scale,
            _logger=_logger
        )

    # Generate output HDUList with masks
    args = inspect.signature(compute_crmasks).parameters
    if crmethod == 'lacosmic':
        prefix_of_excluded_args = 'mm_'
    elif crmethod == 'mmcosmic':
        prefix_of_excluded_args = 'la_'
    else:
        prefix_of_excluded_args = 'xxx'
    filtered_args = {k: v for k, v in locals().items() if
                     k in args and
                     k not in ['list_arrays'] and
                     k[:3] != prefix_of_excluded_args}
    hdu_primary = fits.PrimaryHDU()
    hdu_primary.header['UUID'] = str(uuid.uuid4())
    for i, fluxf in enumerate(flux_factor):
        hdu_primary.header[f'FLUXF{i+1}'] = fluxf
    hdu_primary.header.add_history(f"CRMasks generated by {__name__}")
    hdu_primary.header.add_history(f"at {datetime.now().isoformat()}")
    for key, value in filtered_args.items():
        if isinstance(value, np.ndarray):
            if np.unique(value).size == 1:
                value = value.flatten()[0]
            elif value.ndim == 1 and len(value) == num_images:
                value = str(value.tolist())
            else:
                value = f'array_shape: {value.shape}'
        elif isinstance(value, list):
            value = str(value)
        hdu_primary.header.add_history(f"- {key} = {value}")
    # Include extension with pixels to be replaced by the median value around them
    # (stored as a binary table with four columns: 'X_pixel', 'Y_pixel', 'X_width', 'Y_width')
    if pixels_to_be_replaced_by_local_median is not None:
        col1 = fits.Column(name='X_pixel', format='K',
                           array=[p[0] for p in pixels_to_be_replaced_by_local_median])
        col2 = fits.Column(name='Y_pixel', format='K',
                           array=[p[1] for p in pixels_to_be_replaced_by_local_median])
        col3 = fits.Column(name='X_width', format='K',
                           array=[p[2] for p in pixels_to_be_replaced_by_local_median])
        col4 = fits.Column(name='Y_width', format='K',
                           array=[p[3] for p in pixels_to_be_replaced_by_local_median])
        hdu_table = fits.BinTableHDU.from_columns([col1, col2, col3, col4], name='RPMEDIAN')
        list_hdu_masks.append(hdu_table)

    hdul_masks = fits.HDUList([hdu_primary] + list_hdu_masks)
    return hdul_masks
